/* -*- Mode:C++ -*- */

/* bison-to-bnf convertor */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define true (0==0)
#define false (!true)

#define FN "config.y"
#define EMPTY "{empty}"

void addchar(int ch, char **buf, int *cnt, int *avl)
{
    if (*cnt>=*avl)
    {
	*avl = (*avl>0) ? (*avl*2) : (128);
	*buf = (char*) realloc(*buf, *avl);
	if (*buf==NULL)
	{
	    printf("out of memory.\n");
	    exit(1);
	};
    };

    (*buf)[(*cnt)++]=ch;
}

void addstr(char *x, char **buf, int *cnt, int *avl)
{
    while (*x!=0) addchar(*x++,buf,cnt,avl);
}

char *extract(FILE *s)
{
    char *buf=NULL;
    int avl=0, cnt=0;
    int cur=fgetc(s), depth=0, prev=0;

    /* read preamble */
    while ( ( (prev!=cur) || (cur!='%') ) && (feof(s)==0) )
    {
	prev=cur;
	cur=fgetc(s);
    };

    /* skip second '%' */
    prev=cur; cur=fgetc(s);

    /* read body of rules until '%%' */
    while ( ( (prev!=cur) || (cur!='%') ) && (feof(s)==0) )
    {
	if (cur=='{') depth++;
	if (depth==0) addchar(cur,&buf,&cnt,&avl);
	if (cur=='}') depth--;
	prev=cur;
	cur=fgetc(s);
    };

    /* close buffer */
    addchar(0,&buf,&cnt,&avl);

    return buf;
}

int isws(int x)
{
    int res=false;
    switch(x) { case ' ': case '\t': case '\n': res=true; default: break; };
    return res;
}

void eatws(char **buf)
{
    while (isws((*buf)[1])==true) (*buf)++;
}

char *strip(FILE *s)
{
    int cnt=0, avl=0, show=1;
    char *extracted=extract(s);
    char *tmp=extracted, *res=NULL;
    while (*tmp!=0)
    {
	if (memcmp(tmp,"/*",2)==0) show=0; /* could crash ... */
	if (show==1) addchar(*tmp,&res,&cnt,&avl);
	if (memcmp(tmp,"*/",2)==0) { show=1; tmp+=2; }; /* could crash ... */
	tmp++;
    };
    addchar(0,&res,&cnt,&avl);
    free(extracted);
    return res;
}

char *process(FILE *s)
{
    static int spaces=0;
    char *stripped=strip(s);
    int cnt=0, avl=0, empty=true;
    char *tmp=stripped, *res=NULL;

    /* skip whitespaces at beginning of buffer*/
    while (isws(*tmp)==true) tmp++;

    while (*tmp!=0)
    {
	if (isws(*tmp)==true)
	{
	    if (spaces++==0) addchar(' ',&res,&cnt,&avl);
	}
	else
	{
	    spaces=0;
	    if (*tmp==':')
	    {
		addchar(' ',&res,&cnt,&avl);
		addchar(':',&res,&cnt,&avl);
		addchar(':',&res,&cnt,&avl);
		addchar('=',&res,&cnt,&avl);
		addchar('\n',&res,&cnt,&avl);
		addchar('\t',&res,&cnt,&avl);
		addchar(' ',&res,&cnt,&avl);
		addchar(' ',&res,&cnt,&avl);
		empty=true;
	    }
	    else if (*tmp=='|')
	    {
		if (empty==true) addstr(EMPTY,&res,&cnt,&avl);
		addchar('\n',&res,&cnt,&avl);
		addchar('\t',&res,&cnt,&avl);
		addchar(*tmp,&res,&cnt,&avl);
		addchar(' ',&res,&cnt,&avl);
		empty=true;
	    }
	    else if (*tmp==';')
	    {
		if (empty==true) addstr(EMPTY,&res,&cnt,&avl);
		addchar('\n',&res,&cnt,&avl);
		addchar(*tmp,&res,&cnt,&avl);
		addchar('\n',&res,&cnt,&avl);
		addchar('\n',&res,&cnt,&avl);
		eatws(&tmp);
	    }
	    else
	    {
		empty=false;
		addchar(*tmp,&res,&cnt,&avl);
	    };
	};
	tmp++;
    };
    addchar(0,&res,&cnt,&avl);
    free(stripped);
    return res;
}

int main(int argc, char **argv)
{
    FILE *s=fopen(FN,"r");
    if (s==NULL)
    {
	printf("cannot open '%s' for reading.\n",FN);
    }
    else
    {
	char *out=process(s);
	printf("%s",out);
	free(out);
	fclose(s);
    };
    return 0;
}








