/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str, provides interface */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "cleanup.h"

/* ---------------------------------------------------------------------- */

void process_collector_sources(user *users)
{
    /* --- process any complete input line --- */
    source *tmp=current_sources;
    while (tmp!=NULL)
    {
	if ((tmp->buffer)[0]!=0)
	{
	    char *line;
	    do
	    {
		int l=strlen(tmp->buffer);
		line=extract_first_line(tmp->buffer,&l);
		if (line!=NULL)
		{
		    log((7,NULL,"tmp->state=%s",(tmp->state==sstate_unauthenticated)?("unauth"):("auth")));

		    if (tmp->state==sstate_unauthenticated)
		    {
			authenticate_partner(line,tmp,users);
		    }
		    else
		    {
			/* sender is on stdout */
			fprintf(stdout,"%s",line);
			if (ferror(stdout)!=0)
			{
			    main_fd=-1;
			    log((0,NULL,"got error during write to output stream (stdout)."));
			    /* no test for space: I got it from here originally */
			    sprintf(tmp->buffer+strlen(tmp->buffer),"%s",line);
			};
		    };
		    /* extract_line has a private static buffer */
		};
	    }
	    while ( (line!=NULL) && (main_fd>-1) );
	};
	tmp=tmp->next;
    };
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    char rstate[128];
    user *users=NULL;
    char *log_file="collector.log";
    struct sockaddr_in client;
    int input_socket=-1, new_socket=-1;
    interface *source_if=NULL;
    char *config_file="collector.conf";
    int input_timeout=CPU_SAVER;

    initstate(time(NULL),rstate,128);

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("collector");

    log((2,NULL,"*** collector startup ***"));

    setvbuf(stdout,NULL,_IONBF,0);
    main_fd=fileno(stdout);

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_collector);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    input_timeout=c->input_timeout;
	    cleanup_users(&users); users=c->users; c->users=NULL;
	    if (source_if!=NULL) free_interface(source_if);
	    source_if=new_interface(c->src->ip_address,c->src->port);
	    if ( (source_if->port!=c->src->port) ||
		 (strcmp(source_if->ip_address,c->src->ip_address)!=0) )
	    {
		free_interface(source_if);
		source_if->port=c->src->port;
		source_if->ip_address=sstrdup(c->src->ip_address);
		input_socket=shutdown_socket(input_socket,"from",source_if);
	    };
	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	if ( (input_socket==-1) && (source_if!=NULL) && (source_if->ip_address!=NULL) )
	    input_socket=re_initialise_input(source_if);

	new_socket=process_sources(input_timeout,&input_socket,&client);
	if (new_socket>-1)
	{
	    source *tmp=new_source(st_collector,sockaddr_to_ascii(&client));
	    tmp->src=new_socket;
	    tmp->next=current_sources;
	    current_sources=tmp;
	    send_challenge(tmp);
	    tmp=NULL;
	};

	cleanup_sources();
	process_collector_sources(users);

	if (main_fd==-1) end=true;
    };

    log((7,NULL,"collector shutdown initiated."));

    shutdown_sources();
    shutdown_socket(input_socket,"from",source_if);

    cleanup_users(&users);
    if (source_if!=NULL) free_interface(source_if);

    log((2,NULL,"*** collector shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
