/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>

#include "log.h"
#include "tools.h"

#include "destination.h" /* conf-tools.h needs "destination" */

#include "conf-tools.h"

/* ---------------------------------------------------------------------- */

role_member *new_role_member(char *name, role_member *next)
{
    role_member *res=(role_member *)smalloc(sizeof(role_member));
    res->name=sstrdup(name);
    res->next=next;
    return res;
}

role_member *free_role_member(role_member *x)
{
    role_member *res=NULL;
    if (x!=NULL)
    {
	if (x->name!=NULL) { sfree(x->name); x->name=NULL; };
	res=x->next;
	sfree(x);
    };
    return res;
}

void cleanup_role_members(role_member **role_members)
{
    while ((*role_members)!=NULL)
	(*role_members)=free_role_member(*role_members);
}

role *new_role(char *name, role_member *members, role *next)
{
    role *res=(role *)smalloc(sizeof(role));
    res->name=sstrdup(name);
    res->members=members;
    res->next=next;
    return res;
}

role *free_role(role *x)
{
    role *res=NULL;
    if (x!=NULL)
    {
	if (x->name!=NULL) { sfree(x->name); x->name=NULL; };
	while (x->members!=NULL) x->members=free_role_member(x->members);
	res=x->next;
	sfree(x);
    };
    return NULL;
}

void cleanup_roles(role **roles)
{
    while ((*roles)!=NULL)
	(*roles)=free_role(*roles);
}

/* ---------------------------------------------------------------------- */

sensor *free_sensor(sensor *list)
{
    sensor *res=NULL;

    if (list!=NULL)
    {
	if (list->name!=NULL) sfree(list->name); list->name=NULL;
	if (list->command!=NULL) sfree(list->command); list->command=NULL;
	if (list->parameter!=NULL) sfree(list->parameter); list->parameter=NULL;
	res=list->next;
	sfree(list);
    };

    return res;
}

sensor *int_add_sensor(sensor *list, int freq,
		       char *name, char *command,
		       char *parameter, const char *fn, int ln)
{
    sensor *x=(sensor *)smalloc(sizeof(sensor));

    x->name=int_sstrdup(name,fn,ln);
    x->frequency=freq;
    x->command=int_sstrdup(command,fn,ln);
    x->parameter=int_sstrdup((parameter!=NULL)?(parameter):(""),fn,ln);
    x->next=list;

    return x;
}

/* ---------------------------------------------------------------------- */

interface *int_new_interface(char *ip, int port, const char *fn, int ln)
{
    interface *res=(interface *)int_smalloc(sizeof(interface),fn,ln);
    res->port=port;
    res->ip_address=sstrdup(ip);
    return res;
}

void int_free_interface(interface *i, const char *fn, int ln)
{
    i->port=0;
    sfree(i->ip_address); i->ip_address=NULL;
    sfree(i);
}

/* ---------------------------------------------------------------------- */

out_dependency *new_dependency(out_dependency *next, char *table, char *signature)
{
    out_dependency *res=(out_dependency *)smalloc(sizeof(out_dependency));
    res->next=next;
    res->table=sstrdup(table);
    res->signature=sstrdup(signature);
    return res;
}

out_dependency *free_dependency(out_dependency *x)
{
    out_dependency *res=x->next;
    x->next=NULL;
    if (x->table!=NULL) sfree(x->table);
    if (x->signature!=NULL) sfree(x->signature);
    sfree(x);
    return res;
}

void cleanup_output_dependencies(out_dependency **x)
{
    while (*x!=NULL) *x=free_dependency(*x);
}

out_struct *new_output(out_struct *next, char *name, char *role, char *heading, out_type tpe, unsigned long delay, unsigned long max_lines, char *query)
{
    out_struct *res=smalloc(sizeof(out_struct));
    res->next=next;
    res->name=sstrdup(name);
    res->role=sstrdup(role);
    res->heading=sstrdup(heading);
    res->tpe=tpe;
    res->delay=delay;
    res->max_lines=max_lines;
    res->min_timestamp=((unsigned long)(-1));
    res->dependencies=NULL;
    res->query=sstrdup(query);
    res->data=NULL;
    res->data_timestamp=0;
    return res;
}

out_struct *free_output(out_struct *x)
{
    out_struct *res=x->next;

    x->next=NULL;
    if (x->name!=NULL) { sfree(x->name); x->name=NULL; };
    if (x->role!=NULL) { sfree(x->role); x->role=NULL; };
    if (x->heading!=NULL) { sfree(x->heading); x->heading=NULL; };
    while (x->dependencies!=NULL)
	x->dependencies=free_dependency(x->dependencies);
    if (x->query!=NULL) { sfree(x->query); x->query=NULL; };
    if (x->data!=NULL) { sfree(x->data); x->data=NULL; };
    sfree(x);

    return res;
}

void cleanup_outputs(out_struct **x)
{
    while (*x!=NULL) *x=free_output(*x);
}

/* ---------------------------------------------------------------------- */

user *new_user(char *name, char *secret, user *next)
{
    user *res=(user*)smalloc(sizeof(user));
    res->name=sstrdup(name);
    res->secret=sstrdup(secret);
    res->next=next;
    return res;
}

user *free_user(user *u)
{
    user *res=NULL;
    if (u!=NULL)
    {
	res=u->next;
	if (u->name!=NULL) sfree(u->name); u->name=NULL;
	if (u->secret!=NULL) sfree(u->secret); u->secret=NULL;
	sfree(u);
    };
    return res;
}

void cleanup_users(user **users)
{
    while ((*users)!=NULL)
	(*users)=free_user(*users);
}

/* ---------------------------------------------------------------------- */

viewer_input *new_viewer_input(char *name, unsigned long rows, char *file, viewer_input *next)
{
    viewer_input *res=(viewer_input*)smalloc(sizeof(viewer_input));
    res->name=sstrdup(name);
    res->file=sstrdup(file);
    res->rows=rows;
    res->next=next;
    return res;
}

viewer_input *free_viewer_input(viewer_input *x)
{
    viewer_input *res=NULL;
    if (x!=NULL)
    {
	if (x->name!=NULL) { sfree(x->name); x->name=NULL; };
	if (x->file!=NULL) { sfree(x->file); x->file=NULL; };
	x->rows=0;
	res=x->next;
	sfree(x);
    };
    return res;
}

void cleanup_viewer_inputs(viewer_input **viewer_inputs)
{
    while ((*viewer_inputs)!=NULL)
	(*viewer_inputs)=free_viewer_input(*viewer_inputs);
}


/* ---------------------------------------------------------------------- */

config *new_config()
{
    config *res=(config *)smalloc(sizeof(config));

    res->log_file=NULL;

    res->db_host=NULL;
    res->db_name=NULL;
    res->db_user=NULL;
    res->db_password=NULL;

    res->src=NULL;
    res->dst=NULL;

    res->identity=NULL;
    res->users=NULL;
    res->roles=NULL;

    res->config_type=ct_unknown;

    res->site_name=NULL;
    res->machine_name=NULL;

    res->receiver_password=NULL;

    res->sensor_list=NULL;
    res->outputs=NULL;
    res->destination_list=NULL;

    res->viewer_inputs=NULL;

    res->input_timeout=0;
    res->directory_frequency=0;

    res->data_rows=0;
    res->data_freq=0;
    res->data_name=NULL;
    res->data_file=NULL;

    return res;
}

void free_config(config *c)
{
    if (c!=NULL)
    {
	if (c->log_file!=NULL) { sfree(c->log_file); c->log_file=NULL; };

	if (c->db_host!=NULL) sfree(c->db_host);
	if (c->db_name!=NULL) sfree(c->db_name);
	if (c->db_user!=NULL) sfree(c->db_user);
	if (c->db_password!=NULL) sfree(c->db_password);

	if (c->src!=NULL) free_interface(c->src);
	if (c->dst!=NULL) free_interface(c->dst);

	while (c->identity!=NULL)
	    c->identity=free_user(c->identity);

	while (c->roles!=NULL)
	    c->roles=free_role(c->roles);

	while (c->users!=NULL)
	    c->users=free_user(c->users);

	while (c->sensor_list!=NULL)
	    c->sensor_list=free_sensor(c->sensor_list);

	while (c->outputs!=NULL)
	    c->outputs=free_output(c->outputs);

	while (c->destination_list!=NULL)
	    c->destination_list=free_destination(c->destination_list);

	while (c->viewer_inputs!=NULL)
	    c->viewer_inputs=free_viewer_input(c->viewer_inputs);

	if (c->site_name!=NULL) { sfree(c->site_name); c->site_name=NULL; };
	if (c->machine_name!=NULL) { sfree(c->machine_name); c->machine_name=NULL; };

	if (c->receiver_password!=NULL) { sfree(c->receiver_password); c->receiver_password=NULL; };

	c->data_rows=0;
	c->data_freq=0;
	if (c->data_name!=NULL) { sfree(c->data_name); c->data_name=NULL; };
	if (c->data_file!=NULL) { sfree(c->data_file); c->data_file=NULL; };

	sfree(c);
	c=NULL;
    };
}

/* ---------------------------------------------------------------------- */
