/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

void config_cleanup();

/* ---------------------------------------------------------------------- */

typedef struct _role_member
{
    struct _role_member *next;
    char *name;
} role_member;

role_member *new_role_member(char *name, role_member *next);
role_member *free_role_member(role_member *x);
void cleanup_role_members(role_member **role_members);

typedef struct _role
{
    char *name;
    struct _role *next;
    role_member *members;
} role;

role *new_role(char *name, role_member *members, role *next);
role *free_role(role *x);
void cleanup_roles(role **roles);

/* ---------------------------------------------------------------------- */

typedef struct _sensor
{
    char *name;
    char *command;
    char *parameter;
    int frequency;
    struct _sensor *next;
} sensor;

sensor *free_sensor(sensor *list);
#define add_sensor(l,f,n,c,p) int_add_sensor(l,f,n,c,p,__FILE__,__LINE__)
sensor *int_add_sensor(sensor *list, int freq,
		       char *name, char *command,
		       char *parameter, const char *fn, int ln);

/* ---------------------------------------------------------------------- */

typedef struct
{
    int port;
    char *ip_address;
} interface;
#define new_interface(i,p) int_new_interface(i,p,__FILE__,__LINE__)
interface *int_new_interface(char *ip, int port, const char *fn, int ln);
#define free_interface(i) int_free_interface(i,__FILE__,__LINE__)
void int_free_interface(interface *i, const char *fn, int ln);

/* ---------------------------------------------------------------------- */

typedef enum
{
    out_type_unknown=0,
    out_type_scalar,
    out_type_histogram,
    out_type_table
} out_type;

typedef struct _out_dependency
{
    char *table;
    char *signature;
    struct _out_dependency *next;
} out_dependency;

out_dependency *new_dependency(out_dependency *next, char *table, char *signature);
out_dependency *free_dependency(out_dependency *x);
void cleanup_output_dependencies(out_dependency **x);

typedef struct _out_struct
{
    struct _out_struct *next;
    int available;
    char *name;
    char *role;
    char *heading;
    out_type tpe;
    unsigned long delay;
    unsigned long max_lines;
    unsigned long min_timestamp;
    out_dependency *dependencies;
    char *query;
    char *data;
    unsigned long data_timestamp;
} out_struct;

out_struct *new_output(out_struct *next, char *name, char *role, char *heading, out_type tpe, unsigned long delay, unsigned long max_lines, char *query);
out_struct *free_output(out_struct *x);
void cleanup_outputs(out_struct **x);

/* ---------------------------------------------------------------------- */

typedef struct _user
{
    char *name;
    char *secret;
    struct _user *next;
} user;

user *new_user(char *name, char *secret, user *next);
user *free_user(user *u);
void cleanup_users(user **users);

/* ---------------------------------------------------------------------- */

typedef struct _viewer_input
{
    char *name, *file;
    unsigned long rows;
    struct _viewer_input *next;
} viewer_input;

viewer_input *new_viewer_input(char *name, unsigned long rows, char *file, viewer_input *next);
viewer_input *free_viewer_input(viewer_input *x);
void cleanup_viewer_inputs(viewer_input **x);

/* ---------------------------------------------------------------------- */

typedef enum
{
    ct_unknown=0,
    ct_watcher,
    ct_collector,
    ct_sender,
    ct_receiver,
    ct_refinery,
    ct_server,
    ct_viewer_1,
    ct_viewer_2
} config_type_enum;

/* ---------------------------------------------------------------------- */

typedef struct
{
    int log_level;
    char *log_file;
    interface *src;
    interface *dst;
    role *roles;
    user *users;
    user *identity;
    char *site_name;
    char *machine_name;
    char *receiver_password;
    char *db_host;
    char *db_name;
    char *db_user;
    char *db_password;
    char *data_name;
    char *data_file;
    sensor *sensor_list;
    out_struct *outputs;
    destination *destination_list;
    viewer_input *viewer_inputs;
    unsigned long data_rows;
    unsigned long data_freq;
    unsigned long input_timeout;
    unsigned long directory_frequency;
    config_type_enum config_type;
} config;

config *new_config();
void free_config(config *c);

/* ---------------------------------------------------------------------- */

config *fetch_config(const char *filename, config_type_enum tpe);

/* ---------------------------------------------------------------------- */
