/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "tools.h" /* provides str */
#include "validate.h"
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str/destination, provides interface */
#include "db.h"

#define max(a,b) (((a)>(b))?(a):(b))

/* ---------------------------------------------------------------------- */

char *stripcr(char *x)
{
    int i=strlen(x);
    static char buf[10000];

    sprintf(buf,"%s",x); if (i>1) i-=1; buf[i]=0;

    return buf;
}

/* ---------------------------------------------------------------------- */

table *new_table(char *name, char *signature, char *last, table *next)
{
    table *tmp=(table*)smalloc(sizeof(table));

    strcpy(tmp->last,last);

    tmp->name=sstrdup(name);
    tmp->signature=sstrdup(signature);

    tmp->next=next;

    return tmp;
}

table *free_table(table *t)
{
    table *res=NULL;

    if (t!=NULL)
    {
	sfree(t->name);
	sfree(t->signature);
	res=t->next;
	sfree(t);
    };

    return res;
}

/* ---------------------------------------------------------------------- */

#ifdef LIBPQ
int pqexec(PGconn *conn, char *cmd)
{
    int res=false;
    PGresult *r=PQexec(conn,cmd);
    if (r!=NULL)
    {
	char buffer[10000];
	static unsigned long pgresqry_time=0;
	ExecStatusType t=PQresultStatus(r);
	switch(t)
	{
	case PGRES_COMMAND_OK:
	case PGRES_TUPLES_OK:
	    res=true;
	    break;
	default:
	    strcpy(buffer,stripcr(cmd));
	    log((1,&pgresqry_time,"pgres query '%s' produced '%s'.",buffer,stripcr(PQresultErrorMessage(r))));
	    end=true;
	    break;
	};
	PQclear(r);
    };
    return res;
}
#endif

/* ---------------------------------------------------------------------- */

relation *new_relation()
{
    relation *res=(relation*)smalloc(sizeof(relation));
    res->rows=0;
    res->columns=0;
    res->cells=NULL;
    return res;
}

void free_relation(relation *x)
{
    if (x->cells!=NULL)
    {
	unsigned long i=0, j=(x->rows)*(x->columns);
	while (i<j) sfree((x->cells)[i++]);
	sfree(x->cells); x->cells=NULL;
    };
    sfree(x);
}

#define s read_field_s
static str *read_field_s=NULL;
char *read_field(FILE *x, unsigned long *c, unsigned long *r)
{
    int  ch=fgetc(x);

    (*c)++;
    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    while ( (feof(x)==0) && (ch!='|') && (ch!='\n') )
    {
	char buf[]={ ch, 0 };
	add_str(s,buf);
	ch=fgetc(x);
    };

    if (ch=='\n') { *c=0; (*r)++; };

    return (clen(s)>0)?(cstr(s)):(NULL);
}
#undef s

relation *do_query(dbsession *session, char *cmd, int *ok)
{
    str *s=new_str();
    relation *res=NULL;
#ifdef LIBPQ
    *ok=false;
    log((7,NULL,"do_select session=%08lx",(unsigned long)session));
    if ( (session!=NULL) && (session->pgc!=NULL) )
    {
	PGresult *data;
	add_str(s,"set timezone to gmt; ");
	add_str(s,cmd);
	data=PQexec(session->pgc,cstr(s));
	log((7,NULL,"do_select(%s) PGresult=%08lx",cstr(s),(unsigned long)data));
	if (data!=NULL)
	{
	    if (PQresultStatus(data)==PGRES_TUPLES_OK)
	    {
		int i, r=PQntuples(data), j, c=PQnfields(data);
		log((7,NULL,"do_select status=ok"));
		if ( (r>0) && (c>0) )
		{
		    res=new_relation();
		    res->rows=r;
		    res->columns=c;
		    res->cells=(relation_cell *)smalloc(r*c*sizeof(relation_cell));
		    for (i=0; i<r; i++)
			for (j=0; j<c; j++)
			    (res->cells)[i*c+j]=sstrdup(PQgetvalue(data,i,j));
		};
		*ok=true;
	    }
	    else
	    {
		static unsigned long select_time=0;
		log((1,&select_time,"command '%s' produced '%s'.",cstr(s),stripcr(PQresultErrorMessage(data))));
		*ok=false;
	    };
		PQclear(data);
	};
    };
#else
    FILE *q;
    add_str(s,"./query.sh 'set timezone to gmt; ");
    add_str(s,cmd);;
    add_str(s,"'");
    q=popen(cstr(s),"r");
    if (q==NULL)
    {
	static unsigned long popen_time=0;
	log((1,&popen_time,"command popen('%s') returned NULL.",cstr(s)));
	*ok=false;
    }
    else
    {
	unsigned long c=0, rows=0, cols=0, element=0, elements=0;
	char *field=read_field(q,&c,&rows);
	*ok=true;
	res=new_relation();
	while (field!=NULL)
	{
	    cols=max(c,cols);
	    if (element>=elements)
	    {
		elements=(elements>0)?(2*elements):(128);
		res->cells=(relation_cell *)srealloc(res->cells,elements*sizeof(relation_cell));
	    };
	    (res->cells)[element++]=sstrdup(field);
	    field=read_field(q,&c,&rows);
	};
	res->rows=rows;
	res->columns=cols+1;

	fclose(q);
    };
#endif
    free_str(s);
    return res;
}

/* ---------------------------------------------------------------------- */

table *read_tables(dbsession *session, char *db_user, int *ok)
{
    table *res=NULL;
#ifdef LIBPQ
    *ok=false;
    log((7,NULL,"read_tables session=%08lx",(unsigned long)session));
    if ( (session!=NULL) && (session->pgc!=NULL) )
    {
	PGresult *data=PQexec(session->pgc,"set timezone to gmt; select nme,sig,last from tables;");
	log((7,NULL,"read_tables PGresult=%08lx",(unsigned long)data));
	if (data!=NULL)
	{
	    if (PQresultStatus(data)==PGRES_TUPLES_OK)
	    {
		table **prev=&res;
		int i, r=PQntuples(data);
		log((7,NULL,"read_tables status=ok"));
		for (i=0; i<r; i++)
		{
		    char *nme=PQgetvalue(data,i,0);
		    char *sig=PQgetvalue(data,i,1);
		    char *lst=PQgetvalue(data,i,2);
		    log((8,NULL,"new_table(nme=%s sig=%s lst=%s)",nme,sig,lst));
		    *prev=new_table(nme,sig,lst,NULL);
		    prev=&((*prev)->next);
		};
		PQclear(data);
		*ok=true;
	    }
	    else
	    {
		char buffer[1000];
		static unsigned long select_time=0;
		log((1,&select_time,"command 'select from tables;' produced '%s'.",stripcr(PQresultErrorMessage(data))));
		PQclear(data);
		pqexec(session->pgc,"create table tables ( nme text, sig text, last timestamp );");
		pqexec(session->pgc,"grant select on tables to public;");
		sprintf(buffer,"grant insert on tables to %s;",db_user);
		pqexec(session->pgc,buffer);
		sprintf(buffer,"grant delete on tables to %s;",db_user);
		pqexec(session->pgc,buffer);
		*ok=false;
	    };
	};
    };
#else
    FILE *s=popen("./tables.sh","r");
    if (s==NULL)
    {
	static unsigned long popen_time=0;
	log((1,&popen_time,"command popen('%s') returned NULL.","./tables.sh"));
	*ok=false;
    }
    else
    {
	int ch, off;
	char x[2], last[30];
	table **prev=&res;
	str *name=new_str(), *signature=new_str();

	*ok=true;
	x[1]=0;

	do
	{
	    /* skip leading space */
	    ch=skip_ws(s);

	    /* read the table name from field 'nme' */
	    reset_str(name);
	    while ( (is_ws(ch)==false) && (feof(s)==0) )
	    { x[0]=ch; add_str(name,x); ch=fgetc(s); };

	    /* skip intermediate whitespace */
	    ch=skip_ws(s);

	    /* read the table type from field 'sig' */
	    reset_str(signature);
	    while ( (is_ws(ch)==false) && (feof(s)==0) )
	    { x[0]=ch; add_str(signature,x); ch=fgetc(s); };

	    /* skip intermediate whitespace */
	    ch=skip_ws(s);

	    off=0;
	    while ( (ch!='\n') && (feof(s)==0) && (off<23) )
	    { last[off++]=ch; ch=fgetc(s); }; last[off++]=0;

	    if (feof(s)==0)
	    {
		log((8,NULL,"new_table(%s,%s,%s)",cstr(name),cstr(signature),last));
		*prev=new_table(cstr(name),cstr(signature),last,NULL);
		prev=&((*prev)->next);
	    };
	} while (feof(s)==0);
	free_str(signature);
	free_str(name);
	fclose(s);
    };
#endif
    return res;
}

/* ---------------------------------------------------------------------- */

unsigned long get_next(dbsession *session, int *ok)
{
    unsigned long res=0;

#ifdef LIBPQ
    *ok=false;
    if ( (session!=NULL) && (session->pgc!=NULL) )
    {
	PGresult *data=PQexec(session->pgc,"select nextval('renamed');");
	if (data!=NULL)
	{
	    if (PQresultStatus(data)==PGRES_TUPLES_OK)
	    {
		if (PQntuples(data)!=1)
		{
		    static unsigned long nextvalcnt_time=0;
		    log((1,&nextvalcnt_time,"select nextval('renamed') did not return 1 tuple."));
		}
		else
		{
		    char *val=PQgetvalue(data,0,0);
		    res=atoi(val);
		    *ok=true;
		};
		PQclear(data);
	    }
	    else
	    {
		static unsigned long nextval_time=0;
		log((1,&nextval_time,"command 'select nextval('renamed');' produced '%s': issuing 'create sequence' command.",stripcr(PQresultErrorMessage(data))));
		PQclear(data);
		*ok=pqexec(session->pgc,"create sequence renamed minvalue 1 start 2;");
		res=1;
	    };
	};
    };
#else
    FILE *s=popen("./nextval.sh","r");
    if (s==NULL)
    {
	static unsigned long popen_time=0;
	log((1,&popen_time,"command popen('%s') returned NULL.","./next.sh"));
	*ok=false;
    }
    else
    {
	if (fscanf(s,"%lu",&res)==1) *ok=true;
	fclose(s);
    };
#endif
    log((7,NULL,"get_next(session)=(%d,%lu)",res,*ok));
    return res;
}

/* ---------------------------------------------------------------------- */

char *add_para(str *s, char *sep, char *lbl, char *val)
{
    if (val!=NULL)
    {
	add_str(s,sep);
	add_str(s,lbl);
	add_str(s,"=");
	add_str(s,val);
	sep=" ";
    };
    return sep;
}

dbsession *db_open(char *db_host, char *db_name, char *db_user, char *db_password)
{
    dbsession *res=(dbsession *)smalloc(sizeof(dbsession));
#ifdef LIBPQ
    char *sep="";
    str *conninfo=new_str();
    /* host hostaddr port dbname user password options tty */
    sep=add_para(conninfo,sep,"host",db_host);
    sep=add_para(conninfo,sep,"dbname",db_name);
    sep=add_para(conninfo,sep,"user",db_user);
    sep=add_para(conninfo,sep,"password",db_password);
    res->pgc=PQconnectdb(cstr(conninfo));
    if (res->pgc==NULL)
    {
	static unsigned long dbconnect_time=0;
	log((1,&dbconnect_time,"PQconnectdb returned NULL."));
	sfree(res);
	res=NULL;
    }
    else
    {
	int s=PQstatus(res->pgc);
	if (s!=CONNECTION_OK)
	{
	    static unsigned long dbconnect_time=0;
	    log((1,&dbconnect_time,"PQconnectdb('%s') returned '%s'.",cstr(conninfo),stripcr(PQerrorMessage(res->pgc))));
	    PQfinish(res->pgc);
	    sfree(res);
	    res=NULL;
	}
	else
	{
	    if (pqexec(res->pgc,"set timezone to gmt;")==false)
	    {
		PQfinish(res->pgc);
		sfree(res);
		res=NULL;
	    };
	};
    };
#else
#ifdef PASS_SECRET
    setenv("DB_HOST",db_host,1);
    setenv("DB_NAME",db_name,1);
    setenv("DB_USER",db_user,1);
    if (db_password!=NULL)
	setenv("DB_PASSWORD",db_password,1);
#endif
    res->buffer=new_str();
    add_str(res->buffer,"set timezone to gmt;");
#endif
    return res;
}

int db_send_commands(dbsession *session, str *s, char **errmsg)
{
    int ok=false;
    log((8,NULL,"sql=%s",cstr(s)));
#ifdef LIBPQ
    if ( (session!=NULL) && (session->pgc!=NULL) )
    {
	int end=false;
	while (end==false)
	{
	    if (clen(s)==0)
	    {
		ok=true;
		end=true;
	    }
	    else
	    {
		char *cmd=extract_first_line(cstr(s),&(s->cnt));
		log((9,NULL,"extract_first_line='%s'",cmd));
		if (pqexec(session->pgc,cmd)==false) end=true;
		log((9,NULL,"after pqexec"));
	    };
	};
    };
#else
    log((7,NULL,"db_send_commands add_str"));
    if ( (session!=NULL) && (session->buffer!=NULL) )
	add_str(session->buffer,cstr(s));
#endif
    log((7,NULL,"db_send_commands OK"));
    return ok;
}

void db_close(dbsession *session)
{
    if (session!=NULL)
    {
#ifdef LIBPQ
	if (session->pgc!=NULL)
	{
	    PQfinish(session->pgc);
	    session->pgc=NULL;
	};
#else
	if (session->buffer!=NULL)
	{
	    FILE *d=popen("./updates.sh","w");
	    if (d==NULL)
	    {
		static unsigned long popen_time=0;
		log((1,&popen_time,"command popen('%s') returned NULL.","./updates.sh"));
	    }
	    else
	    {
		fprintf(d,"%s\\q\n",cstr(session->buffer)); fclose(d);
	    };
	    free_str(session->buffer);
	    session->buffer=NULL;
	};
#endif
	sfree(session);
    };
}

/* ---------------------------------------------------------------------- */

void db_cleanup()
{
    if (read_field_s!=NULL) { free_str(read_field_s); read_field_s=NULL; };
}

/* ---------------------------------------------------------------------- */
