/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <netdb.h>
#include <netinet/in.h>

#include <bzlib.h>

#ifdef BZIP_OLD
#define COMPRESS bzBuffToBuffCompress
#else
#define COMPRESS BZ2_bzBuffToBuffCompress
#endif

#include "log.h"
#include "tools.h"
#include "deliver.h"

/* ---------------------------------------------------------------------- */

static char *extract_buffer=NULL;
int extract_data(destination *d, char **outbuf, int *outlen)
{
    int res=false;
    static int avl=0;

    *outlen=0;
    *outbuf=NULL;

    if (clen(d->s)!=0)
    {
	char *first=cstr(d->s), *last=rindex(first,'\n')+1;

	if (last==NULL)
	{
	    log((0,NULL,"reached sendsize=%d, but couldn't find <LF> in buffer '%s'.",d->sendsize,first));
	}
	else
	{
	    unsigned int srclen=last-first;
	    if (avl<srclen+1)
	    {
		avl=srclen+1;
		extract_buffer=(char*)srealloc(extract_buffer,avl);
	    };
	    memcpy(extract_buffer,first,srclen);
	    extract_buffer[srclen]=0;
	    res=true;
	    *outlen=srclen;
	    *outbuf=extract_buffer;
	    memcpy(cstr(d->s),cstr(d->s)+srclen,clen(d->s)-srclen);
	    d->s->cnt -= srclen;
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

static char *compress_buffer=NULL;
int compress_data(char *srcbuf, int srclen, char **outbuf, int *outlen)
{
    int res=false;
    static int avl=0;
    unsigned int rc, dstlen=(3*srclen)/2+2*sizeof(long);

/*
    log((7,NULL,"!!!ignoring compression")); *outbuf=srcbuf; *outlen=srclen; return true;
*/

    if (avl<dstlen)
    {
	avl=dstlen;
	compress_buffer=(char*)srealloc(compress_buffer,avl);
    };

    ((long *)(compress_buffer))[0]=random();
    ((long *)(compress_buffer))[1]=random();

    rc=COMPRESS(compress_buffer+2*sizeof(long),&dstlen,srcbuf,srclen,9,0,30);
    if (rc==BZ_OK)
    {
	res=true;
	*outbuf=compress_buffer;
	*outlen=dstlen+2*sizeof(long);
    }
    else
    {
	static unsigned long compression_time=0;
	log((1,&compression_time,"compression error, rc=%d.",rc));
    };

    return res;
}

int encrypt_data(destination *d,char *buffer, int length)
{
    int res=false;

    if ( (d!=NULL) && (length>0) )
    {
	log((7,NULL,"sender_password='%s'",d->password));
	cbc_encrypt(d->password,buffer,length);
	res=true;
    };
    return res;
}

char *encode_data(char *buffer, int length)
{
    char *res=base64_encode(buffer,length);
    return res;
}

int post_data(destination *d, char *data)
{
    int res=false;
    char dest[10000];
    char *hostname=(d->proxy_name==NULL)?(d->hostname):(d->proxy_name);
    struct hostent *hp=gethostbyname(hostname);

    if (hp==NULL)
    {
	static unsigned long gethostbyname_time=0;
	log((1,&gethostbyname_time,"error locating '%s' through gethostbyname().",hostname));
    }
    else
    {
	main_fd=socket(AF_INET,SOCK_STREAM,0);
	if (main_fd<0)
	{
	    static unsigned long socket_time=0;
	    log((1,&socket_time,"error during socket()."));
	}
	else
	{
	    struct sockaddr_in server;
	    int port=(d->proxy_name==NULL)?(d->port):(d->proxy_port);
	    sprintf(dest,"%s:%d",hostname,port);
	    server.sin_family=AF_INET;
	    memcpy(&(server.sin_addr),(hp->h_addr),hp->h_length);
	    server.sin_port=htons(port);
	    if (connect(main_fd,(const struct sockaddr *)&server,sizeof(server))<0)
	    {
		static unsigned long connect_time=0;
		log((1,&connect_time,"post_data: error during connect(%s:%d).",hostname,port));
		close(main_fd);
		main_fd=-1;
	    }
	    else
	    {
		char num[50];
		int written=0;
		str *s=new_str();

		if (d->proxy_name==NULL)
		{
		    add_str(s,"POST ");
		    add_str(s,d->path);
		    sprintf(dest+strlen(dest),"%s",d->path);
		}
		else
		{
		    char num[10]; sprintf(num,"%d",d->port);
		    add_str(s,"POST http://");
		    add_str(s,d->hostname);
		    add_str(s,":");
		    add_str(s,num);
		    add_str(s,d->path);
		    sprintf(dest+strlen(dest),"|%s:%d%s",d->hostname,d->port,d->path);
		};
		add_str(s," HTTP/1.0\r\n");
		if (d->proxy_name==NULL)
		{
		    add_str(s,"Host: ");
		    add_str(s,d->hostname);
		};
		add_str(s,"Content-Type: application/octet-stream\r\n");
		sprintf(num,"Content-Length: %d\r\n",strlen(data)); add_str(s,num);
		if (d->credentials!=NULL)
		{
		    add_str(s,"Proxy-Authorization: Basic ");
		    add_str(s,d->credentials);
		    add_str(s,"\r\n");
		};
		add_str(s,"\r\n");
		add_str(s,data);
		res=true;
		while (written<clen(s))
		{
		    int rc=write(main_fd,cstr(s)+written,clen(s)-written);
		    if (rc>-1)
		    {
			written+=rc;
		    }
		    else
		    {
			log((0,NULL,"error on write to '%s:%d'.",hostname,port));
			written=clen(s);
			res=false;
		    };
		};
		if (res==true)
		{
		    char tmp[1000], *code;
		    int rc=1;
		    reset_str(s);
		    while (rc>0)
		    {
			rc=read(main_fd,tmp,999);
			if (rc>-1)
			{
			    tmp[rc]=0;
			    add_str(s,tmp);
			}
			else
			{
			    log((0,NULL,"error on read from '%s:%d'.",hostname,port));
			    res=false;
			};
		    };
		    log((8,NULL,"HTTP answer for '%s'=%s",dest,cstr(s)));
		    code=index(cstr(s),' ');
		    if (code==NULL)
		    {
			static unsigned long postspace_time=0;
			log((1,&postspace_time,"HTTP POST answer for '%s' did not contain a space.",dest));
			res=false;
		    }
		    else
		    {
			int o=0;
			char cb[10];

			code++;
			while ( (isdigit(*code)) && (o<10) ) cb[o++]=*code++;
			cb[o++]=0;

			if (strcmp(cb,"200")!=0)
			{
			    static unsigned long postcode_time=0;
			    log((1,&postcode_time,"HTTP POST answer for '%s' returned %s (expected 200).",dest,cb));
			    res=false;
			};
		    };
		};
		close(main_fd);
		main_fd=-1;
		free_str(s);
	    };
	};
    };

    return res;
}

void deliver_block(destination *d)
{
    int re_insert=false;
    int extracted_length=0;
    char *extracted_data=NULL;

    if (extract_data(d,&extracted_data,&extracted_length)==true)
    {
	int compressed_length=0;
	char *compressed_data=NULL;
	char *hex=tohex(extracted_data,extracted_length);

	log((9,NULL,"extracted=%s",extracted_data));
	log((7,NULL,"after  extract   ='%s'",hex)); sfree(hex);
	if (compress_data(extracted_data,extracted_length,
			  &compressed_data,&compressed_length)==false)
	{
	    static unsigned long compress_time=0;
	    log((1,&compress_time,"compress data ilen=%d failed.",extracted_length));
	    re_insert=true;
	}
	else
	{
	    char *hex;

	    /* my cbc-encrypter needs a multiple of 8, so round up */
	    compressed_length+=7;
	    compressed_length&=0xfffffff8;

	    hex=tohex(compressed_data,compressed_length);
	    log((7,NULL,"after  compress  ='%s'",tohex(compressed_data,compressed_length))); sfree(hex);

	    if (encrypt_data(d,compressed_data,compressed_length)==false)
	    {
		static unsigned long encrypt_time=0;
		log((1,&encrypt_time,"encrypt data ilen=%d failed.",compressed_length));
		re_insert=true;
	    }
	    else
	    {
		char *encoded=encode_data(compressed_data,compressed_length);
		char *hex=tohex(compressed_data,compressed_length);
		log((7,NULL,"after  encrypt   ='%s'",hex)); sfree(hex);
		if ( (encoded==NULL) || (*encoded==0) )
		{
		    static unsigned long encode_time=0;
		    log((1,&encode_time,"base64 encode ilen=%d failed.",compressed_length));
		    re_insert=true;
		}
		else
		{
		    log((8,NULL,"encoded=%s",extracted_data));

		    if (post_data(d,encoded)==false)
		    {
			static unsigned long post_time=0;
			log((1,&post_time,"post_data failed."));
			re_insert=true;
		    };
		    /* "encoded" is a static buffer and doesn't need freeing */
		};
		/* "cbc_encrypt" modifies in-place so nothing needs freeing */
	    };
	    /* "compressed" is a static buffer and doesn't need freeing */
	};
	/* "extracted" is a static buffer and doesn't need freeing */
    };

    if (re_insert==true)
    {
	/* no test for space: I got it from here in the first place */
	add_str(d->s,extracted_data);
    };
}

/* ---------------------------------------------------------------------- */

void deliver_cleanup()
{
    if (extract_buffer!=NULL)
    { sfree(extract_buffer); extract_buffer=NULL; };
    if (compress_buffer!=NULL)
    { sfree(compress_buffer); compress_buffer=NULL; };
}

/* ---------------------------------------------------------------------- */
