/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

#include "log.h"
#include "tools.h"

#define LFNSIZE 1000

static int log_level=DEFAULT_LOG_LEVEL;
static char log_file[LFNSIZE];
static char log_application[LFNSIZE];

/* ---------------------------------------------------------------------- */

void log_setlevel(int level)
{
    if ( (level>=0) && (level<=9) ) log_level=level;
}

/* ---------------------------------------------------------------------- */

void log_setapp(const char *application)
{
    snprintf(log_application,LFNSIZE,"%s",application);
}

/* ---------------------------------------------------------------------- */

void log_setfile(const char *file)
{
    snprintf(log_file,LFNSIZE,"%s",file);
}

/* ---------------------------------------------------------------------- */

FILE *log_openfile()
{
    FILE *res=NULL;
    res=fopen(log_file,"a");
    if (res==NULL)
    {
	char buf[LFNSIZE];
	snprintf(buf,LFNSIZE,
		 "/tmp/%s_help_me_%d.log",log_application,getpid());
	log_setfile(buf);
	res=fopen(log_file,"a");
    };
    return res;
}

/* ---------------------------------------------------------------------- */

int log_avl=0;
unsigned char *log_buf=NULL;

void resize_log_buf(int sz)
{
    if (log_avl<=sz)
    {
	log_avl=sz+1;
	log_buf=(char*)srealloc(log_buf,log_avl);
    };
}

void int_log(int msg_level, time_t *lastmsg, const char *fmt, ...)
{
    int do_log=false;
    char *add_msg=NULL;
    unsigned long now=time(NULL);

    switch(msg_level)
    {
    case 0:
	/*
	  - n=0: error messages (loss of I/O, write errors, configuration
	  parsing errors) are logged
	*/
	do_log=true; if (lastmsg!=NULL) *lastmsg=now;
	break;

    case 1:
	/*
	  - n=1: protocol, memory and data errors are logged once daily
	  - n=4: protocol, memory, data errors are logged once hourly
	*/
	if ( (log_level>=1) && (now-(*lastmsg)>=24*60*60) )
	{ do_log=true; *lastmsg=now; add_msg="[suspending this message for 1 day.]"; };
	if ( (log_level>=4) && (now-(*lastmsg)>=60*60) )
	{ do_log=true; *lastmsg=now; add_msg="[suspending this message for 1 hour.]"; };
	if (log_level>=5) { do_log=true; *lastmsg=now; add_msg=NULL; };
	break;

    case 2:
	/*
	  - n=2: normal events (start/shutdown, reconfiguration,
	  start/shutdown of connections) are logged
	*/
	if (log_level>=2) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;

    case 3:
	/*
	  - n=3: messages describing configuration data are emitted
	*/
	if (log_level>=3) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;

    case 6:
	/*
	  - n=6: input/output meta-data is logged (length, names)
	*/
	if (log_level>=6) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;

    case 7:
	/*
	  - n=7: internal states and actions are logged (alloc list,
	  execution points)
	*/
	if (log_level>=7) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;

    case 8:
	/*
	  - n=8: all inputs and outputs are logged
	*/
	if (log_level>=8) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;

    default:
	/*
	  - n=9: internal data is logged also (no, IBM don't pay me
	  comission on hdd sales!)
	*/
	if (log_level>=9) { do_log=true; if (lastmsg!=NULL) *lastmsg=now; };
	break;
    };

    if (do_log==true)
    {
	unsigned char limit[1000];
	unsigned char *limit_buf=limit, *print_buf=NULL;
	FILE *l=log_openfile();
	if (l!=NULL)
	{
	    int i, len, cnt=10000;

	    if (log_unlimited==true)
	    {
		do
		{
		    va_list arg;
		    va_start(arg, fmt);
		    if (log_avl<cnt) resize_log_buf(cnt+1);
		    cnt=vsnprintf(log_buf, log_avl, fmt, arg);
		    va_end(arg);
		}
		while (cnt>=log_avl);
	    }
	    else
	    {
		va_list arg;

		va_start(arg, fmt);
		vsnprintf(limit_buf, 1000, fmt, arg);
		va_end(arg);
	    };

	    if (add_msg!=NULL)
	    {
		if (log_unlimited==true)
		{
		    len=strlen(log_buf);
		    resize_log_buf(len+strlen(add_msg)+2);
		    sprintf(log_buf+len," %s",add_msg);
		}
		else
		{
		    len=strlen(limit_buf);
		    if (len<1000)
			snprintf(limit_buf+len,1000-len," %s",add_msg);
		};
	    };

	    print_buf=(log_unlimited==true)?(log_buf):(limit_buf);

	    for (i=0,len=strlen(print_buf); i<len; i++)
	    {
		if (print_buf[i]=='\t')
		{
		    print_buf[i]=32; /* space */
		}
		else if (print_buf[i]=='\n')
		{
		    print_buf[i]=171; /* double arrow */
		}
		else if ( (print_buf[i]<32) || (print_buf[i]>127) )
		{
		    print_buf[i]=183; /* middle dot */
		};
	    };

	    if (log_level<7)
	    {
		fprintf(l,"%s %d %s\n",iso_timestamp(),msg_level,print_buf);
	    }
	    else
	    {
		fprintf(l,"%s %d %s (%s[%d])\n",iso_timestamp(),msg_level,print_buf,log_filename,log_linenumber);
	    };

	    fclose(l);
	};
    };
}

/* ---------------------------------------------------------------------- */

void log_cleanup()
{
    log_avl=0;
    if (log_buf!=NULL) { sfree(log_buf); log_buf=NULL; }
}

/* ---------------------------------------------------------------------- */
