#
# sysmon tree
#

proc ReadTables {dbc} {
    pg_select $dbc "SELECT * from tables order by last;" array {
	puts [format "%s %s %s" $array(nme) $array(sig) $array(last)]
    }
}

proc RunBrowser {} {
    # Create a TixTree widget to display the hypothetical DOS disk drive
    #
    tixTree .mytree  -options {
	hlist.separator "/"
	hlist.width 35
	hlist.height 25
    }

    pack .mytree -expand yes -fill both -padx 10 -pady 10 -side left

    set hlist [.mytree subwidget hlist]

    .mytree config -opencmd "DynTree:OpenDir .mytree"

    # Add the root directory the TixTree widget
    DynTree:AddDir .mytree /
}

proc DynTree:AddDir {tree dir} {
    set hlist [$tree subwidget hlist]
    if {$dir == "/"} { set text / } else { set text [file tail $dir] }
    $hlist add $dir -itemtype imagetext -text $text -image [tix getimage folder]
    $tree setmode $dir open
}

# This command is called whenever the user presses the (+) indicator or
# double clicks on a directory whose mode is "open". It loads the files
# inside that directory into the Tree widget.
#
proc DynTree:OpenDir {tree dir} {
    set hlist [$tree subwidget hlist]

    if {[$hlist info children $dir] != {}} {
	foreach kid [$hlist info children $dir] {
	    $hlist show entry $kid
	}
	return
    }

    set files [list "a" "b" "c"]

    foreach f $files {
	if {$f == "a"} {
	    if {$dir == "/"} {
		set subdir /$f
	    } else {
		set subdir $dir/$f
	    }
	    DynTree:AddDir $tree $subdir
	} else {
	    if {$dir == "/"} {
		set file /$f
	    } else {
		set file $dir/$f
	    }

	    $hlist add $file -itemtype imagetext -text $f -image [tix getimage file]
	}
    }
}

proc Finish {} {
    global $dbc
    pg_disconnect $dbc
    vwait .
    exit
}

load libpgtcl[info sharedlibextension]
set dbc [pg_connect "sysmon" -host "192.168.2.85"]
RunBrowser
bind . <Destroy> Finish

