/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <ctype.h> /* provides isalpha/isdigit */
#include <stdio.h> /* provides FILE */
#include <string.h> /* provides strncmp */
#include <unistd.h> /* provides write */

#include "log.h"
#include "tools.h" /* needs FILE, provides str */
#include "destination.h" /* needs str, provides destination */
#include "conf-tools.h" /* needs destination, provides out_struct */
#include "source.h" /* needs str/interface, provides read_line */
#include "db.h" /* provides tables */
#include "output.h" /* needs str/tables/out_struct */

/* ---------------------------------------------------------------------- */

int eat_tab(char **buf)
{
    int res=false;
    if (**buf=='\t')
    {
	(*buf)++;
	res=true;
    };
    return res;
}

int issymbol(char x)
{
    if (isalpha(x)) return true;
    if (isdigit(x)) return true;
    switch(x)
    {
    case '_':
    case ':':
	return true;
    default:
	break;
    };
    return false;
}

out_struct *make_output(str *line, out_struct *next)
{
    static unsigned long dir_parse=0;

    char save[1000];
    char *name=NULL;
    char *role=NULL;
    char *heading=NULL;
    int available=false;
    char *tmp=cstr(line);
    int ok=eat_tab(&tmp);
    out_struct *res=NULL;
    out_type tpe=out_type_unknown;
    unsigned long delay=0, max_lines=0, last=0;

    log((8,NULL,"make_output(%s);",cstr(line)));

    strncpy(save,tmp,999); save[999]=0;

    if (ok==true)
    {
	name=tmp;
	while (issymbol(*tmp)==true) tmp++;
	if (strlen(name)==0) ok=false;
	if (ok==false) log((1,&dir_parse,"can't read name in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { *tmp=0; tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	role=tmp;
	while (issymbol(*tmp)==true) tmp++;
	if (strlen(role)==0) ok=false;
	if (ok==false) log((1,&dir_parse,"can't read role in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { *tmp=0; tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	switch(*tmp)
	{
	case 't': available=true; break;
	case 'f': available=false; break;
	default: ok=false; break;
	};
	if (ok==false) log((1,&dir_parse,"can't read availability in '%s'.",save));
	tmp++;
    };
    if (ok==true) { if (*tmp=='\t') { tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	if (sscanf(tmp,"%lu",&last)==1)
	{
	    while (isdigit(*tmp)) tmp++;
	}
	else
	{
	    ok=false;
	};
	if (ok==false) log((1,&dir_parse,"can't read min_timestamp in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	ok=false;

	if (strncmp(tmp,"scalar",6)==0)
	{ ok=true; tmp+=6; tpe=out_type_scalar; };

	if (strncmp(tmp,"histogram",9)==0)
	{ ok=true; tmp+=9; tpe=out_type_histogram; };

	if (strncmp(tmp,"table",5)==0)
	{ ok=true; tmp+=5; tpe=out_type_table; };

	if (ok==false) log((1,&dir_parse,"can't read type in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	if (sscanf(tmp,"%lu",&delay)==1)
	{
	    while (isdigit(*tmp)) tmp++;
	}
	else
	{
	    ok=false;
	};
	if (ok==false) log((1,&dir_parse,"can't read delay in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	if (sscanf(tmp,"%lu",&max_lines)==1)
	{
	    while (isdigit(*tmp)) tmp++;
	}
	else
	{
	    ok=false;
	};
	if (ok==false) log((1,&dir_parse,"can't read max_lines in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\t') { tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	heading=tmp;
	while (issymbol(*tmp)==true) tmp++;
	if (strlen(heading)==0) ok=false;
	if (ok==false) log((1,&dir_parse,"can't read heading in '%s'.",save));
    };
    if (ok==true) { if (*tmp=='\n') { *tmp=0; tmp++; } else { ok=false; }; };
    if (ok==true)
    {
	res=new_output(next, name, role, heading, tpe, delay, max_lines, "");
	res->available=available;
	res->min_timestamp=last;
    };
    if (ok==false) { log((1,&dir_parse,"parse error in '%s'.",save)); };

    return res;
}

out_struct *get_outputs(socket_buf *sb, int *ocnt)
{
    int ok, sock=*(sb->socket);
    str *tmp=new_str();
    out_struct *res=NULL;

    *ocnt=0;

    add_str(tmp,"directory\n");
    log((9,NULL,"get_outputs 0 server_socket=%d.",sb->socket));
    ok=write_string(tmp,sock);
    log((9,NULL,"get_outputs 1 server_socket=%d.",sb->socket));
    if ( (ok!=true) && (*(sb->socket)>-1) )
    {
	close(*(sb->socket));
	*(sb->socket)=-1;
    };

    flush_buf(sb);
    log((9,NULL,"get_outputs 2 server_socket=%d.",sb->socket));
    ok=read_line(sb,tmp);
    log((9,NULL,"get_outputs 3 server_socket=%d.",sb->socket));

    if (strncmp("\tdirectory\t",cstr(tmp),11)!=0)
    {
	static unsigned long proto_time=0;
	log((1,&proto_time,"answer for 'directory' cmd was '%s': closing socket %d.",cstr(tmp),*(sb->socket)));

	if (*(sb->socket)>-1)
	{
	    close(*(sb->socket));
	    *(sb->socket)=-1;
	};
    }
    else
    {
	int i,j;
	*ocnt=atoi(cstr(tmp)+11);
	for (i=0, j=0; j<*ocnt; j++)
	{
	    out_struct *next=NULL;
	    ok=read_line(sb,tmp);
	    if (ok==false) reset_str(tmp);
	    next=make_output(tmp,res);

	    if (next==NULL)
	    {
		*ocnt=0;
		cleanup_outputs(&res);
	    }
	    else
	    {
		res=next;
	    };
	};
    };

    free_str(tmp);

    return res;
}

/* ---------------------------------------------------------------------- */

int fetch_output(socket_buf *sb, out_struct *asked)
{
    char buf[100], *num;
    str *tmp=new_str(), *data=new_str();
    int ok, lines, res=true, sock=*(sb->socket);

    log((7,NULL,"fetch_output(%s)",asked->name));

    if (asked->data!=NULL) { sfree(asked->data); asked->data=NULL; };

    add_str(tmp,"\t");
    add_str(tmp,asked->name);
    sprintf(buf,"\t%lu\n",asked->max_lines); add_str(tmp,buf);

    log((7,NULL,"sending to server: '%s'",cstr(tmp)));

    ok=write_string(tmp,sock);
    if ( (ok!=true) && (*(sb->socket)>-1) )
    {
	close(*(sb->socket));
	*(sb->socket)=-1;
    };

    flush_buf(sb);
    ok=read_line(sb,tmp);

    log((7,NULL,"line from server: '%s'",cstr(tmp)));

    if (strncmp("\tdirectory\t",cstr(tmp),11)==0) res=false;

    num=index(cstr(tmp),'\t');
    if (num==NULL) { res=false; } else { num=index(num+1,'\t'); };
    if (num==NULL)
    {
	res=false;
    }
    else
    {
	if (sscanf(num+1,"%d",&lines)!=1) res=false;
    };

    if (res==true)
    {
	int i=0;
	add_str(data,cstr(tmp));
	while (i++<lines)
	{
	    ok=read_line(sb,tmp);
	    add_str(data,cstr(tmp));
	    if (ok==false) { res=false; i=lines; };
	};
    };

    if (res==true) asked->data=sstrdup(cstr(data));

    free_str(data);
    free_str(tmp);

    return res;
}

/* ---------------------------------------------------------------------- */

out_struct *compute_availability(table *tables, out_struct *outputs, unsigned long *ocnt, char *request)
{
    out_struct *res=NULL;

    *ocnt=0;
    while (outputs!=NULL)
    {
	int available=true;
	out_dependency *depend=outputs->dependencies;

	(*ocnt)++;
	while (depend!=NULL)
	{
	    int found=false;
	    table *tmp=tables;

	    while (tmp!=NULL)
	    {
		if (strcmp(tmp->name,depend->table)==0)
		{
		    if (strcmp(tmp->signature,depend->signature)==0)
		    {
			unsigned long ts=iso_to_epoch(tmp->last);
			if (outputs->min_timestamp>ts)
			    outputs->min_timestamp=ts;
			found=true;
		    }
		    else
		    {
			found=false;
		    };
		    tmp=NULL;
		}
		else
		{
		    tmp=tmp->next;
		};
	    };
	    if (found==false)
	    {
		depend=NULL;
		available=false;
	    }
	    else
	    {
		depend=depend->next;
	    };
	};
	if (request!=NULL) if (strcmp(request,outputs->name)==0) res=outputs;
	outputs->available=available;
	outputs=outputs->next;
    };

    return res;
}

/* ---------------------------------------------------------------------- */

int permission(char *rname, role *roles, char *user)
{
    int res=false;

    if ( (roles==NULL) || (user==NULL) ) return true;

    while (roles!=NULL)
    {
	if (strcmp(rname,roles->name)==0)
	{
	    role_member *m=roles->members;
	    while (m!=NULL)
	    {
		if (strcmp(user,m->name)==0)
		{
		    res=true;
		    m=NULL;
		}
		else
		{
		    m=m->next;
		};
	    };
	    roles=NULL;
	}
	else
	{
	    roles=roles->next;
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

void produce_directory(out_struct *outputs, unsigned long ocnt, str *s, role *roles, char *user, int use_permissions)
{
    char header[100];
    reset_str(s);

    sprintf(header,"\tdirectory\t%lu\n",ocnt);
    add_str(s,header);

    while (outputs!=NULL)
    {
	if ( (use_permissions==false) || (permission(outputs->role,roles,user)==true) )
	{
	    char buf[100];
	    static unsigned long type_time=0;

	    /* name */
	    add_str(s,"\t");
	    add_str(s,outputs->name);

	    /* role */
	    add_str(s,"\t");
	    add_str(s,outputs->role);

	    /* available */
	    add_str(s,"\t");
	    add_str(s,(outputs->available==true)?("t"):("f"));

	    /* minimum timestamp */
	    sprintf(buf,"\t%lu",outputs->min_timestamp);
	    add_str(s,buf);

	    /* display type */
	    switch(outputs->tpe)
	    {
	    case out_type_scalar: add_str(s,"\tscalar"); break;
	    case out_type_histogram: add_str(s,"\thistogram"); break;
	    case out_type_table: add_str(s,"\ttable"); break;
	    default: add_str(s,"\ttable");
		log((1,&type_time,"outputs contained unknown type %d.",outputs->tpe)); break;
	    };

	    /* recommended delay */
	    sprintf(buf,"\t%lu",outputs->delay);
	    add_str(s,buf);

	    /* recommended delay */
	    sprintf(buf,"\t%lu",outputs->max_lines);
	    add_str(s,buf);

	    /* heading */
	    add_str(s,"\t");
	    add_str(s,outputs->heading);

	    add_str(s,"\n");
	};
	outputs=outputs->next;
    };
}

/* ---------------------------------------------------------------------- */





