/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <bzlib.h>

#ifdef BZIP_OLD
#define DECOMPRESS bzBuffToBuffDecompress
#else
#define DECOMPRESS BZ2_bzBuffToBuffDecompress
#endif

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str */

#include "db.h"
#include "log.h"
#include "cache.h"
#include "cleanup.h"
#include "validate.h"

/* ---------------------------------------------------------------------- */

char *skip_field(char *data)
{
    int i=0;
    while (i<5)
    {
	if (data!=NULL) data=index(data,'\t')+1;
	i++;
    };
    if (data!=NULL) data--;
    return data;
}

/* ---------------------------------------------------------------------- */

char *split_record(char *data, v_type *type, str *title, str *units, str *value)
{
    char x[2]; x[1]=0;

    data++;
    *type=vt_unknown;
    if (memcmp(data,"int",3)==0)
    {
	*type=vt_int;
	data+=3;
    }
    else if (memcmp(data,"float",5)==0)
    {
	*type=vt_float;
	data+=5;
    }
    else if (memcmp(data,"timestamp",9)==0)
    {
	*type=vt_timestamp;
	data+=9;
    }
    else if (memcmp(data,"text",4)==0)
    {
	*type=vt_text;
	data+=4;
    };

    data++; while (*data!='\t')
    { x[0]=*data++; add_str(title,x); };

    data++; while (*data!='\t')
    { x[0]=*data++; add_str(units,x); };

    data++; while ( (*data!='\t') && (*data!='\n') )
    { x[0]=*data++; add_str(value,x); };

    return data;
}

/* ---------------------------------------------------------------------- */

str *extract_type_title=NULL;
#define title extract_type_title
str *extract_type_units=NULL;
#define units extract_type_units
str *extract_type_value=NULL;
#define value extract_type_value

void extract_type(char *data, str *signature)
{
    int i=0;
    char *sep="";
    while (i<4) { data=skip_field(data); i++; };

    while (*data!=0x0a)
    {
	v_type type;
	char number[10];

	if (title==NULL) { title=new_str(); } else { reset_str(title); };
	if (units==NULL) { units=new_str(); } else { reset_str(units); };
	if (value==NULL) { value=new_str(); } else { reset_str(value); };

	data=split_record(data,&type,title,units,value);

	add_str(signature,sep);
	sprintf(number,"%d",type); add_str(signature,number);
	add_str(signature,"/");
	add_str(signature,cstr(title));
	add_str(signature,"/");
	add_str(signature,cstr(units));

	sep=":";
    };
}
#undef title
#undef units
#undef value

/* ---------------------------------------------------------------------- */

str *sql_new_name_res=NULL;
#define res sql_new_name_res
char *sql_new_name(dbsession *session, config *c, char *name)
{
    int ok=false;
    char *result=NULL;
    static char buf[100];
    unsigned long num=get_next(session,&ok);
    if (num>0)
    {
	if (res==NULL) { res=new_str(); } else { reset_str(res); };
	add_str(res,name); sprintf(buf,"_%04lx",num); add_str(res,buf);
	result=cstr(res);
    };
    return result;
}
#undef res

/* ---------------------------------------------------------------------- */

str *sql_rename_res=NULL;
#define res sql_rename_res
char *sql_rename(dbsession *session, config *c, char *name)
{
    char *result=NULL;
    char *new_name=sql_new_name(session,c,name);
    if (new_name!=NULL)
    {
	if (res==NULL) { res=new_str(); } else { reset_str(res); };

	add_str(res,"alter table ");
	add_str(res,name);
	add_str(res," rename to ");
	add_str(res,new_name);
	add_str(res,";\ndelete from tables where nme='");
	add_str(res,name);
	add_str(res,"';\n");

	result=cstr(res);
    };
    return result;
}
#undef res

/* ---------------------------------------------------------------------- */

str *sql_create_res=NULL;
#define res sql_create_res
char *sql_create(char *name, char *signature, char *db_user)
{
    if (res==NULL) { res=new_str(); } else { reset_str(res); };

    add_str(res,"insert into tables values ( '");
    add_str(res,name);
    add_str(res,"', '");
    add_str(res,signature);
    add_str(res,"', '2000-01-01 00:00:00-00');\ncreate table ");
    add_str(res,name);
    add_str(res," ( ts timestamp");

    log((7,NULL,"signature=%s",signature));

    while (*signature!=0)
    {
	char *beg=signature+2;
	char *end=index(beg,'/');

	add_str(res,", ");

	if (end==NULL) end=beg+strlen(beg)-1;
	if (end<beg) end=beg;

	cpy_str(res,beg,end-beg);

	switch(*signature)
	{
	case '1':
	    add_str(res," integer");
	    break;
	case '2':
	    add_str(res," float");
	    break;
	case '3':
	    add_str(res," timestamp");
	    break;
	case '4':
	    add_str(res," text");
	    break;
	default:
	    log((0,NULL,"internal error: signature 0."));
	    break;
	};
	beg=index(beg,':');
	signature=(beg==NULL)?(signature+strlen(signature)):(beg+1);
    };
    add_str(res," );\n");

    add_str(res,"grant select on ");
    add_str(res,name);
    add_str(res," to public;\n");

    add_str(res,"grant insert on ");
    add_str(res,name);
    add_str(res," to ");
    add_str(res,db_user);
    add_str(res,";\n");

    log((7,NULL,"create cmd=%s",cstr(res)));

    return cstr(res);
}
#undef res

/* ---------------------------------------------------------------------- */

str *sql_insert_res=NULL;
#define res sql_insert_res
str *sql_insert_title=NULL;
#define title sql_insert_title
str *sql_insert_units=NULL;
#define units sql_insert_units
str *sql_insert_value=NULL;
#define value sql_insert_value
char *sql_insert(char *name, char *data)
{
    int i=0;
    char *sep="";
    if (res==NULL) { res=new_str(); } else { reset_str(res); };

    add_str(res,"insert into ");
    add_str(res,name);
    add_str(res," values ( ");

    while (i<3) { data=skip_field(data); i++; };

    while (*data!=0x0a)
    {
	v_type type;
	static unsigned long inserttype_time=0;

	if (title==NULL) { title=new_str(); } else { reset_str(title); };
	if (units==NULL) { units=new_str(); } else { reset_str(units); };
	if (value==NULL) { value=new_str(); } else { reset_str(value); };

	data=split_record(data,&type,title,units,value);

	switch(type)
	{
	case vt_int:
	case vt_float:
	    add_str(res,sep);
	    add_str(res,cstr(value));
	    break;
	case vt_timestamp:
	case vt_text:
	    add_str(res,sep);
	    add_str(res,"'");
	    add_str(res,cstr(value));
	    add_str(res,"'");
	    break;
	default:
	    log((1,&inserttype_time,"invalid type during 'insert' construction."));
	    break;
	};
	sep=", ";
    };
    add_str(res," );\n");

    return cstr(res);
}
#undef res
#undef title
#undef units
#undef value

/* ---------------------------------------------------------------------- */

str *sql_update_res=NULL;
#define res sql_update_res
char *sql_update(char *name, char *last)
{
    log((8,NULL,"sql_update(%s,%s)",name,last));

    if (res==NULL) { res=new_str(); } else { reset_str(res); };

    add_str(res,"update tables set last='");
    add_str(res,last);
    add_str(res,"' where nme='");
    add_str(res,name);
    add_str(res,"';\n");

    return cstr(res);
}
#undef res

/* ---------------------------------------------------------------------- */

char *decode_input(char *data, int *length)
{
    char *res=base64_decode(data,length);
    sfree(data);
    return res;
}

/* ---------------------------------------------------------------------- */

void decrypt_input(char *data, int length, char *password)
{
    log((7,NULL,"receiver_password='%s'",password));
    cbc_decrypt(password, data, length);
}

/* ---------------------------------------------------------------------- */

char *unpack_input(char *data, int *length)
{
    char *res=NULL;
    int mylen=(*length)-2*sizeof(long);
    unsigned int dstlen=(*length)*80;
    char *dstbuf=(char*)smalloc(dstlen);
    int rc=DECOMPRESS(dstbuf,&dstlen,data+2*sizeof(long),mylen,0,0);
    char *hex=tohex(data,mylen);
    log((7,NULL,"before decompress='%s'",hex)); sfree(hex);
    if (rc==BZ_OK)
    {
	res=dstbuf;
    }
    else
    {
	log((0,NULL,"decompression error %d.",rc));
	sfree(dstbuf);
    };
    /* base64_decode maintains a static buffer, so don't free it */
    return res;
}

/* ---------------------------------------------------------------------- */

char *read_input()
{
    char *res=NULL;
    char *query_method=getenv("REQUEST_METHOD");

    if ( (query_method==NULL) || (strcmp(query_method,"POST")!=0) )
    {
	static unsigned long protocoltype_time=0;
	log((1,&protocoltype_time,"expected 'POST', but got '%s'.\n",query_method));
    }
    else
    {
	char *lenstr=getenv("CONTENT_LENGTH");
	if (lenstr==NULL)
	{
	    static unsigned long contentlength_time=0;
	    log((1,&contentlength_time,"could not retrieve 'CONTENT_LENGTH'.\n"));
	}
	else
	{
	    int len=atoi(lenstr);
	    res=(char*)smalloc(len+1);
	    memset(res,0,len+1);
	    fread(res,len,1,stdin);
	};
    };
    return res;
}

/* ---------------------------------------------------------------------- */

void receiver_cleanup()
{
#define CLEANUP(X) if (X!=NULL) { free_str(X); X=NULL; }
    CLEANUP(sql_new_name_res);
    CLEANUP(extract_type_title);
    CLEANUP(extract_type_units);
    CLEANUP(extract_type_value);
    CLEANUP(sql_rename_res);
    CLEANUP(sql_create_res);
    CLEANUP(sql_insert_res);
    CLEANUP(sql_insert_title);
    CLEANUP(sql_insert_units);
    CLEANUP(sql_insert_value);
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    config *c;
    char *errmsg=NULL;
    char *log_file="receiver.log";
    char *config_file="receiver.conf";

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("receiver");

    log((2,NULL,"*** receiver startup ***"));

    c=fetch_config(config_file,ct_receiver);
    if (c==NULL)
    {
	errmsg="couldn't read config file.";
    }
    else
    {
	int length=0;
	char *data, *off=NULL, *decoded=NULL, *unpacked=NULL;

	log_setfile(c->log_file);
	log_setlevel(c->log_level);

	log((7,NULL,"before read_input()"));
	data=read_input();

	log((8,NULL,"data=%s",data));

	if (data==NULL)
	{
	    errmsg="couldn't read input.\n";
	}
	else
	{
	    log((7,NULL,"before decode_input()"));
	    decoded=decode_input(data,&length);
	};

	if (decoded==NULL)
	{
	    errmsg="couldn't decode input.\n";
	}
	else
	{
	    char *hex=tohex(decoded,length);
	    log((7,NULL,"before decrypt_input()"));
	    log((7,NULL,"before decrypt   ='%s'",hex)); sfree(hex);
	    decrypt_input(decoded,length,c->receiver_password);
	};

	if (decoded!=NULL)
	{
	    char *hex;
	    log((7,NULL,"before unpack_input()"));
/*
	    log((7,NULL,"!!!ignoring decompression")); unpacked=decoded;
*/
	    unpacked=unpack_input(decoded,&length);
	    sfree(decoded); /* newly added, does this work? */
	    hex=tohex(unpacked,strlen(unpacked));
	    log((7,NULL,"before usage     ='%s'",hex)); sfree(hex);
	};

	if (unpacked==NULL)
	{
	    errmsg="couldn't unpack input.\n";
	}
	else
	{
	    dbsession *session;

	    log((9,NULL,"unpacked=%s",unpacked));
	    log((7,NULL,"before db_open(%s,%s,%s,%s)",c->db_host,c->db_name,c->db_user,c->db_password));
	    session=db_open(c->db_host,c->db_name,c->db_user,c->db_password);
	    if (session==NULL)
	    {
		errmsg="couldn't open database connection.\n";
	    }
	    else
	    {
		int ok;
		table *tabs;

		log((7,NULL,"before read_tables()"));
		tabs=read_tables(session,c->db_user,&ok);
		if (ok==false)
		{
		    errmsg="couldn't read table 'tables': create command issued, please try again.";
		}
		else
		{
		    char *cmd;
		    int txn=false;
		    str *s=new_str();
		    str *name=new_str();
		    str *signature=new_str();
		    str *ts=new_str();
		    str *site=new_str();
		    str *machine=new_str();
		    str *sensor=new_str();
		    table **prv=NULL, *tmp=NULL, *match=NULL;

		    data=unpacked;
		    off=index(data,'\n');

		    log((7,NULL,"before while(off!=NULL)"));
		    while (off!=NULL)
		    {
			int rc, t=off[1]; off[1]=0;

			reset_str(ts);
			reset_str(site);
			reset_str(machine);
			reset_str(sensor);

			log((8,NULL,"line=%s",data));

			rc=validate_line(data,site,machine,sensor,ts);
			if (rc==true)
			{
			    int found=false;

			    log((6,NULL,"validate=%s",(rc==true)?("true"):("false")));

			    reset_str(name);
			    add_str(name,cstr(site));
			    add_str(name,"_");
			    add_str(name,cstr(machine));
			    add_str(name,"_");
			    add_str(name,cstr(sensor));

			    reset_str(signature);
			    extract_type(data,signature);

			    log((9,NULL,"signature=%s",cstr(signature)));

			    prv=&tabs;
			    match=NULL;
			    while (*prv!=NULL)
			    {
				log((7,NULL,"beginning of while;"));
				if (strcmp(cstr(name),(*prv)->name)==0)
				{
				    log((7,NULL,"table '%s' exists.",cstr(name)));
				    found=true;
				    match=*prv;
				    if (strcmp(cstr(signature),(*prv)->signature)!=0)
				    {
					char *rename=sql_rename(session,c,cstr(name));
					log((7,NULL,"record signature (%s) differs from table '%s' signature (%s).",cstr(signature),cstr(name),(*prv)->signature));
					if (rename==NULL)
					{
					    log((7,NULL,"subroutine failure: not executing rename operation."));
					}
					else
					{
					    log((7,NULL,"executing rename operation."));
					    log((9,NULL,"rename command=%s",rename));
					    if (txn==true) { add_str(s,"commit;\n"); txn=false; };
					    add_str(s,rename);
					};
					log((7,NULL,"tmp=*prv;"));
					tmp=*prv;
					log((7,NULL,"*prv=(*prv)->next;"));
					*prv=(*prv)->next;
					log((7,NULL,"free_table(%08lx);",(unsigned long)tmp));
					free_table(tmp);
					log((7,NULL,"match=NULL;"));
					match=NULL;
					log((7,NULL,"found=false;"));
					found=false;
				    }
				    else
				    {
					match=*prv;
					prv=&((*prv)->next);
					log((7,NULL,"signatures (%s) match.",cstr(signature)));
				    };
				}
				else
				{
				    prv=&((*prv)->next);
				};
				log((7,NULL,"end of while;"));
			    };
			    log((7,NULL,"loop finished."));
			    log((9,NULL,"match=%08lx",(unsigned long)match));
			    if (found==false)
			    {
				log((7,NULL,"creating new table '%s'.",cstr(name)));
				if (txn==true) { add_str(s,"commit;\n"); txn=false; };
				cmd=sql_create(cstr(name),cstr(signature),c->db_user);
				log((9,NULL,"adding command '%s'.",cmd));
				add_str(s,cmd);
				match=new_table(cstr(name),cstr(signature),"0000-00-00 00:00:00-00",tabs);
				tabs=match;
			    };
			    if (txn==false) { add_str(s,"begin;\n"); txn=true; };
			    log((7,NULL,"create test finished."));
			    log((9,NULL,"record.last=%s",cstr(ts)));
			    log((9,NULL,"match->last=%s",(match==NULL)?("*NULL"):(match->last)));
			    if ( (match!=NULL) && (memcmp(cstr(ts),match->last,23)>0) )
			    {
				memcpy(match->last,cstr(ts),23);
				log((6,NULL,"sensor '%s' gets last='%s'.",match->name,match->last));
			    };
			    cmd=sql_insert(cstr(name),data);
			    log((9,NULL,"adding command '%s'.",cmd));
			    add_str(s,cmd);
			};
			off[1]=t;
			data=off+1;
			off=index(data,'\n');
		    };
		    log((7,NULL,"after while(off!=NULL)"));
		    log((7,NULL,"before update-command loop."));
		    tmp=tabs;
		    while (tmp!=NULL)
		    {
			if (txn==false) { add_str(s,"begin;\n"); txn=true; };
			cmd=sql_update(tmp->name,tmp->last);
			log((9,NULL,"adding command '%s'.",cmd));
			add_str(s,cmd);
			tmp=tmp->next;
		    };
		    if (txn==true) { add_str(s,"commit;\n"); txn=false; };
		    log((7,NULL,"before db_send_commands()."));
		    db_send_commands(session,s,&errmsg);
		    log((7,NULL,"after db_send_commands()."));

		    log((7,NULL,"before free_str(name)."));
		    free_str(name);
		    log((7,NULL,"before free_str(signature)."));
		    free_str(signature);
		    log((7,NULL,"before free_str(ts)."));
		    free_str(ts);
		    log((7,NULL,"before free_str(site)."));
		    free_str(site);
		    log((7,NULL,"before free_str(machine)."));
		    free_str(machine);
		    log((7,NULL,"before free_str(sensor)."));
		    free_str(sensor);
		    log((7,NULL,"before free_str(s)."));
		    free_str(s);

		    log((7,NULL,"before sfree(unpacked)."));
/*
		    log((7,NULL,"!!! ignoring sfree(unpacked)."));
*/
		    sfree(unpacked);
		};
		log((7,NULL,"before free_table loop."));
		while (tabs!=NULL)
		    tabs=free_table(tabs);
	    };
	    log((7,NULL,"before db_close(session)."));
	    db_close(session);
	    log((7,NULL,"after db_close(session)."));
	};
	free_config(c);
    };

    log((7,NULL,"producing outputs."));
    if (errmsg==NULL)
    {
	printf("Status: 200 OK\r\n");
    }
    else
    {
	unsigned long dummy=0;
	log((1,&dummy,"returning status 500 %s",errmsg));
	printf("Status: 500 %s\r\n",errmsg);
    };

    printf("\r\n");
    printf("And thanks for all the fish.\r\n");

    fflush(stdout);

    db_cleanup();
    receiver_cleanup();
    validate_cleanup();

    log((2,NULL,"*** receiver shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
