/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str/destination, provides interface/out_struct */

#include "db.h" /* needs str */
#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "cleanup.h"
#include "output.h" /* needs out_struct */

static str *produce_x_line_s=NULL;
#define s produce_x_line_s
void produce_output(out_struct *my_output, unsigned long count, dbsession *session, int socket)
{
    relation *tab;
    int r, c, ok=false;
    char *tmp=(char*)smalloc(30+strlen(my_output->query));

    sprintf(tmp,my_output->query,count);
    tab=do_query(session,tmp,&ok);

    if (s==NULL) { s=new_str(); } else { reset_str(s); };
    sprintf(tmp,"\t%s\t%lu",my_output->name,(ok==true)?(tab->rows):(0)); add_str(s,tmp);
    add_str(s,"\n");

    if (ok==true)
    {
	for (r=0; r<tab->rows; r++)
	{
	    for (c=0; c<tab->columns; c++)
	    {
		add_str(s,"\t");
		add_str(s,(tab->cells)[r*(tab->columns)+c]);
	    };
	    add_str(s,"\n");
	};
    };

    if (tab!=NULL) free_relation(tab);

    sfree(tmp);

    write_string(s,socket);
}
#undef s

int process_command(char *db_host, char *db_name, char *db_user, char *db_pswd, char *cmd, int socket, out_struct *outputs)
{
    int ok, res=socket;
    char *request=NULL;
    unsigned long ocnt, count=0;
    out_struct *my_output;
    dbsession *session=db_open(db_host,db_name,db_user,db_pswd);
    table *tables=read_tables(session,db_user,&ok);

    if (cmd!=NULL) if (cmd[0]=='\t') request=cmd+1;
    if (request!=NULL)
    {
	cmd=index(request,'\t');
	if (cmd!=NULL) *cmd++=0;
	if (cmd!=NULL) sscanf(cmd,"%lu",&count);
    };

    my_output=compute_availability(tables, outputs, &ocnt, request);

    if (my_output==NULL)
    {
	str *s=new_str();
	produce_directory(outputs,ocnt,s,NULL,NULL,true);
	write_string(s,socket);
	free_str(s);
    }
    else
    {
	produce_output(my_output,(count>my_output->max_lines)?
		       (my_output->max_lines):(count),session,socket);
    };

    while (tables!=NULL) tables=free_table(tables);

    db_close(session);
    return res;
}

/* ---------------------------------------------------------------------- */

void process_refinery_source(char *db_host, char *db_name, char *db_user, char *db_pswd, user *users, out_struct *outputs)
{
    char *line=NULL;

    /* --- process any complete input line --- */
    source *tmp=current_sources;
    while (tmp!=NULL)
    {
	if ((tmp->buffer)[0]!=0)
	{
	    do
	    {
		int l=strlen(tmp->buffer);
		line=extract_first_line(tmp->buffer,&l);
		if (line!=NULL)
		{
		    if (tmp->state==sstate_unauthenticated)
		    {
			authenticate_partner(line,tmp,users);
		    }
		    else
		    {
			tmp->src=process_command(db_host, db_name, db_user, db_pswd, line, tmp->src, outputs);
		    };
		};
	    }
	    while ( (line!=NULL) && (tmp->src>-1) );
	};
	tmp=tmp->next;
    };
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    user *users=NULL;
    char *db_host=NULL;
    char *db_name=NULL;
    char *db_user=NULL;
    char *db_pswd=NULL;
    out_struct *outputs=NULL;
    interface *source_if=NULL;
    struct sockaddr_in client;
    int input_timeout=CPU_SAVER;
    int input_socket=-1, new_socket=-1;

    char *log_file="refinery.log";
    char *config_file="refinery.conf";

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("refinery");

    log((2,NULL,"*** refinery startup ***"));

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_refinery);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    input_timeout=c->input_timeout;
	    cleanup_users(&users); users=c->users; c->users=NULL;
	    if (db_host!=NULL) sfree(db_host); db_host=sstrdup(c->db_host);
	    if (db_name!=NULL) sfree(db_name); db_name=sstrdup(c->db_name);
	    if (db_user!=NULL) sfree(db_user); db_user=sstrdup(c->db_user);
	    if (db_pswd!=NULL) sfree(db_pswd); if (c->db_password!=NULL) db_pswd=sstrdup(c->db_password);
	    if (source_if!=NULL) free_interface(source_if);
	    source_if=new_interface(c->src->ip_address,c->src->port);
	    if ( (source_if->port!=c->src->port) ||
		 (strcmp(source_if->ip_address,c->src->ip_address)!=0) )
	    {
		free_interface(source_if);
		source_if->port=c->src->port;
		source_if->ip_address=sstrdup(c->src->ip_address);
		input_socket=shutdown_socket(input_socket,"from",source_if);
	    };
	    cleanup_outputs(&outputs); outputs=c->outputs; c->outputs=NULL;
	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	if ( (input_socket==-1) && (source_if!=NULL) && (source_if->ip_address!=NULL) )
	    input_socket=re_initialise_input(source_if);

	new_socket=process_sources(input_timeout,&input_socket,&client);
	if (new_socket>-1)
	{
	    if (current_sources==NULL)
	    {
		current_sources=new_source(st_sender,"server-socket");
		current_sources->src=new_socket;
		send_challenge(current_sources);
	    }
	    else
	    {
		log((2,NULL,"accepted unexpected connection from '%s:%d', closing it.",ulong_to_dotted(client.sin_addr.s_addr),client.sin_port));
		close(new_socket);
	    };
	};

	if ( (db_host!=NULL) && (db_name!=NULL) && (db_user!=NULL) )
	    process_refinery_source(db_host, db_name, db_user,
				    db_pswd, users, outputs);

	cleanup_sources();
    };

    log((7,NULL,"refinery shutdown initiated."));

    shutdown_sources();
    shutdown_socket(input_socket,"from",source_if);

    log((2,NULL,"*** refinery shutdown ***"));

    cleanup_outputs(&outputs);

    cleanup_users(&users);
    if (db_host!=NULL) sfree(db_host); db_host=NULL;
    if (db_name!=NULL) sfree(db_name); db_name=NULL;
    if (db_user!=NULL) sfree(db_user); db_user=NULL;
    if (source_if!=NULL) free_interface(source_if); source_if=NULL;

    /* clean up my static string */
    if (produce_x_line_s!=NULL) free_str(produce_x_line_s);
    produce_x_line_s=NULL;

    /* the database module needs cleaning up */
    db_cleanup();

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
