/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str, provides interface */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "cleanup.h"
#include "deliver.h"
#include "validate.h"

/* ---------------------------------------------------------------------- */

int check_machines(destination *d, char *machine)
{
    int res=false;
    machine_filter *f=d->machine_filters;

    while (f!=NULL)
    {
	if (strcmp(f->machine,machine)==0)
	{
	    res=true;
	    f=NULL;
	}
	else
	{
	    f=f->next;
	};
    };

    return res;
}

int check_sensors(destination *d, char *machine, char *sensor)
{
    int res=false;
    sensor_filter *f=d->sensor_filters;

    while (f!=NULL)
    {
	if ( (strcmp(f->machine,machine)==0) && (strcmp(f->sensor,sensor)==0) )
	{
	    res=true;
	    f=NULL;
	}
	else
	{
	    f=f->next;
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

static str *do_filtering_buf=NULL;
#define buf do_filtering_buf
static str *do_filtering_ts=NULL;
#define ts do_filtering_ts
static str *do_filtering_site=NULL;
#define site do_filtering_site
static str *do_filtering_machine=NULL;
#define machine do_filtering_machine
static str *do_filtering_sensor=NULL;
#define sensor do_filtering_sensor
void do_filtering(char *mysite)
{
    char *line=NULL;

    do
    {
	int len=strlen(current_sources->buffer);

	line=extract_last_line(current_sources->buffer,&len);
	if (line!=NULL)
	{
	    if (buf==NULL) { buf=new_str(); } else { reset_str(buf); };
	    if (ts==NULL) { ts=new_str(); } else { reset_str(ts); };
	    if (site==NULL) { site=new_str(); } else { reset_str(site); };
	    if (machine==NULL) { machine=new_str(); } else { reset_str(machine); };
	    if (sensor==NULL) { sensor=new_str(); } else { reset_str(sensor); };

	    add_str(buf,"\ttext\tsite\tname\t");
	    add_str(buf,mysite);
	    add_str(buf,line);

	    if (validate_line(cstr(buf),site,machine,sensor,ts)==true)
	    {
		destination *tmp=current_destinations;
		while (tmp!=NULL)
		{
		    int insert
			=(check_machines(tmp,cstr(machine))==true)
			?(true)
			:(check_sensors(tmp,cstr(machine),cstr(sensor)));

		    if (insert==true) add_str(tmp->s,cstr(buf));

		    tmp=tmp->next;
		};
	    };
	};
    }
    while (line!=NULL);
}
#undef buf
#undef ts
#undef site
#undef machine
#undef sensor

/* ---------------------------------------------------------------------- */

void process_sender_destinations(destination *tmp)
{
    while (tmp!=NULL)
    {
	unsigned long now=time(NULL);
	log((7,NULL,"checking destination %s:%d%s: len=%d",tmp->hostname,tmp->port,tmp->path,clen(tmp->s)));
	log((7,NULL,"size(buf)-size(sendsize)=%d",clen(tmp->s)-(tmp->sendsize)));
	log((7,NULL,"now-(tmp->lastdelivery)=%d",now-(tmp->lastdelivery)));
	log((7,NULL,"maxwait=%d",(tmp->maxwait)));
	if ( (clen(tmp->s)>tmp->sendsize) ||
	     (now-(tmp->lastdelivery)>tmp->maxwait) )
	{
	    log((7,NULL,"doing delivery."));
	    tmp->lastdelivery=now;
	    deliver_block(tmp);
	};
	tmp=tmp->next;
    };
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    char rstate[128];
    char *config_file="sender.conf";
    char *log_file="sender.log";
    char *site_name=NULL;
    int input_socket=-1;

    initstate(time(NULL),rstate,128);

    parse_commands(argc, argv, &config_file, &log_file);

    log_setapp("sender");
    log_setfile(log_file);

    log((2,NULL,"*** sender startup ***"));

    install_handler();

    current_sources=new_source(st_sender,"stdin");
    current_sources->src=fileno(stdin);
    setvbuf(stdin,NULL,_IONBF,0);

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_sender);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    if (site_name!=NULL)
	    {
		sfree(site_name);
		site_name=NULL;
	    };
	    site_name=c->site_name;
	    c->site_name=NULL;

	    cleanup_destinations();
	    current_destinations=c->destination_list;
	    c->destination_list=NULL;

	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	(void)process_sources(CPU_SAVER,&input_socket,NULL);
	cleanup_sources();
	if (current_sources==NULL)
	{
	    if (end==false)
	    {
		log((0,NULL,"lost input stream (stdin)."));
		end=true;
	    };
	}
	else
	{
	    if (site_name!=NULL)
	    {
		do_filtering(site_name);
		process_sender_destinations(current_destinations);
	    };
	};
    };

    log((7,NULL,"sender shutdown initiated."));

    shutdown_sources();
    cleanup_destinations();

    if (site_name!=NULL) sfree(site_name); site_name=NULL;

    if (do_filtering_buf!=NULL)
    { free_str(do_filtering_buf); do_filtering_buf=NULL; };
    if (do_filtering_ts!=NULL)
    { free_str(do_filtering_ts); do_filtering_ts=NULL; };
    if (do_filtering_site!=NULL)
    { free_str(do_filtering_site); do_filtering_site=NULL; };
    if (do_filtering_sensor!=NULL)
    { free_str(do_filtering_sensor); do_filtering_sensor=NULL; };
    if (do_filtering_machine!=NULL)
    { free_str(do_filtering_machine); do_filtering_machine=NULL; };

    deliver_cleanup();
    validate_cleanup();

    log((2,NULL,"*** clean sender shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
