/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str, provides interface/out_struct */
#include "db.h" /* provides tables */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "output.h" /* needs str/tables/out_struct/socket_buf */
#include "cleanup.h"

/* ---------------------------------------------------------------------- */

int my_outputs_cnt=0;
out_struct *my_outputs=NULL;
out_struct *timed_outputs(socket_buf *sb, int *ocnt, unsigned long freq)
{
    out_struct *res;
    static time_t last=0;
    time_t now=time(NULL);

    if ( (my_outputs==NULL) || (now-last<freq) )
    {
	last=now;
	while (my_outputs!=NULL) my_outputs=free_output(my_outputs);
	my_outputs=get_outputs(sb, &my_outputs_cnt);
    };

    res=my_outputs;
    *ocnt=my_outputs_cnt;

    return res;
}

/* ---------------------------------------------------------------------- */

out_struct *find_output(out_struct *outputs, char *request)
{
    out_struct *res=NULL;
    while (outputs!=NULL) if (strcmp(outputs->name,request)==0)
    {
	res=outputs;
	outputs=NULL;
    }
    else
    {
	outputs=outputs->next;
    };

    return res;
}

/* ---------------------------------------------------------------------- */

int fetch_output_timed(socket_buf *sb, out_struct *asked, str *s)
{
    int res=false;
    unsigned long ts=time(NULL);

    log((7,NULL,"fetch_output_timed(%s)",asked->name));

    if ( (ts-asked->data_timestamp>asked->delay) || (asked->data==NULL) )
	fetch_output(sb,asked);

    reset_str(s);
    if (asked->data!=NULL)
    {
	res=true;
	add_str(s,asked->data);
    };

    return res;
}

/* ---------------------------------------------------------------------- */

static str *fetch_answer_s=NULL;
#define s fetch_answer_s
/* this must check the syntax of the data block! */
char *fetch_answer(char *request, char *user, role *roles, socket_buf *sb, unsigned long freq)
{
    int ocnt=0;
    out_struct *outputs=timed_outputs(sb,&ocnt,freq);

    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    if (outputs==NULL)
    {
	add_str(s,"\tdirectory\t0\n");
    }
    else
    {
	int do_dir=false;
	out_struct *asked=find_output(outputs,request);
	if ( (asked==NULL) || (permission(asked->role,roles,user)==false) )
	{
	    do_dir=true;
	}
	else
	{
	    if (fetch_output_timed(sb,asked,s)==false) do_dir=true;
	};
	if (do_dir==true) produce_directory(outputs,ocnt,s,roles,user,true);

    };

    return cstr(s);
}
#undef s

/* ---------------------------------------------------------------------- */

int process_command(char *cmd, char *user, role *roles, int socket, socket_buf *sb, unsigned long freq)
{
    int res=socket;
    str *s=new_str();
    char *request=NULL, *answer=NULL;
    unsigned long count=0;

    if (cmd!=NULL) if (cmd[0]=='\t') request=cmd+1;
    if (request!=NULL)
    {
	cmd=index(request,'\t');
	if (cmd!=NULL) *cmd++=0;
	if (cmd!=NULL) sscanf(cmd,"%lu",&count);
    };

    answer=
	(request==NULL)
	?fetch_answer("xxx",user,roles,sb,freq)
	:fetch_answer(request,user,roles,sb,freq);

    if (strncmp("\tdirectory\t",answer,11)==0)
    {
	add_str(s,answer);
    }
    else
    {
	char tmp[100];
	int i, len=strlen(answer);
	char *line=extract_first_line(answer,&len);
	char *num=index(index(line,'\t')+1,'\t')+1;
	unsigned long lines=atoi(num);

	*num=0;
	count=(count<lines)?(count):(lines);
	add_str(s,line); sprintf(tmp,"%lu\n",count); add_str(s,tmp);

	for (i=0; i<count; i++)
	{
	    line=extract_first_line(answer,&len);
	    if (line!=NULL)
	    {
		add_str(s,line);
	    }
	    else
	    {
		i=count;
	    };
	};
    };

    log((9,NULL,"process_command: write_string(%s,%d);",cstr(s),socket));

    write_string(s,socket);
    free_str(s);

    return res;
}

/* ---------------------------------------------------------------------- */

void process_server_sources(user *users, role *roles, socket_buf *sb, unsigned long freq)
{
    /* --- process any complete input line --- */
    source *tmp=current_sources;
    while (tmp!=NULL)
    {
	if ((tmp->buffer)[0]!=0)
	{
	    char *line;
	    do
	    {
		int l=strlen(tmp->buffer);
		line=extract_first_line(tmp->buffer,&l);
		if (line!=NULL)
		{
		    log((7,NULL,"tmp->state=%s",(tmp->state==sstate_unauthenticated)?("unauth"):("auth")));

		    if (tmp->state==sstate_unauthenticated)
		    {
			sfree(tmp->t.server.user);
			tmp->t.server.user=authenticate_partner(line,tmp,users);
		    }
		    else
		    {
			log((9,NULL,"calling process_command with server_socket=%d.",sb->socket));
			tmp->src=process_command(line, tmp->t.server.user, roles, tmp->src, sb, freq);
		    };
		    /* extract_line has a private static buffer */
		};
	    }
	    while ( (line!=NULL) && (main_fd>-1) );
	};
	tmp=tmp->next;
    };
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    char rstate[128];
    user *users=NULL;
    user *identity=NULL;
    role *roles=NULL;
    char *log_file="server.log";
    struct sockaddr_in client;
    int input_socket=-1, new_socket=-1;
    interface *refinery_conn=NULL;
    interface *offer_if=NULL;
    char *config_file="server.conf";
    int input_timeout=CPU_SAVER;
    unsigned long freq=30;
    socket_buf *sb=new_socket_buf(&main_fd);

    initstate(time(NULL),rstate,128);

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("server");

    log((2,NULL,"*** server startup ***"));

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_server);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    cleanup_users(&identity); identity=c->identity; c->identity=NULL;

	    if ( (refinery_conn==NULL) ||
		 ( (refinery_conn->port!=c->src->port) ||
		   (strcmp(refinery_conn->ip_address,c->src->ip_address)!=0) ) )
	    {
		if (refinery_conn!=NULL) free_interface(refinery_conn);
		refinery_conn=new_interface(c->src->ip_address,c->src->port);
		if (main_fd!=-1)
		    main_fd=shutdown_socket(main_fd,"to",refinery_conn);
	    };

	    freq=c->directory_frequency;
	    input_timeout=c->input_timeout;
	    cleanup_users(&users); users=c->users; c->users=NULL;
	    cleanup_roles(&roles); roles=c->roles; c->roles=NULL;
	    if (offer_if!=NULL) free_interface(offer_if);
	    offer_if=new_interface(c->dst->ip_address,c->dst->port);
	    if ( (offer_if->port!=c->dst->port) ||
		 (strcmp(offer_if->ip_address,c->dst->ip_address)!=0) )
	    {
		free_interface(offer_if);
		offer_if->port=c->dst->port;
		offer_if->ip_address=sstrdup(c->dst->ip_address);
		input_socket=shutdown_socket(input_socket,"from",offer_if);
	    };
	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	log((7,NULL,"main_fd=%d.",main_fd));
	if (refinery_conn!=NULL)
	    log((7,NULL,"refinery_conn=%s:%d.",refinery_conn->ip_address,refinery_conn->port));
	if (identity!=NULL)
	    log((7,NULL,"identity=%s/%s.",identity->name,identity->secret));
	if ( (main_fd==-1) && (refinery_conn!=NULL) && (identity!=NULL) )
	{
	    log((7,NULL,"attempting to open connection to %s:%d.",refinery_conn->ip_address,refinery_conn->port));

	    main_fd=re_initialise_output(refinery_conn);
	    if (main_fd!=-1)
	    {
		log((7,NULL,"success at opening connection to %s:%d.",refinery_conn->ip_address,refinery_conn->port));
		main_fd=identify_self(main_fd,identity);
	    };
	};

	if ( (input_socket==-1) && (offer_if!=NULL) && (offer_if->ip_address!=NULL) )
	    input_socket=re_initialise_input(offer_if);

	new_socket=process_sources(input_timeout,&input_socket,&client);
	if (new_socket>-1)
	{
	    source *tmp=new_source(st_server,sockaddr_to_ascii(&client));
	    tmp->src=new_socket;
	    tmp->next=current_sources;
	    current_sources=tmp;
	    send_challenge(tmp);
	    tmp->t.server.user=sstrdup("");
	    tmp=NULL;
	};

	cleanup_sources();
	if (*(sb->socket)!=-1) process_server_sources(users,roles,sb,freq);
    };

    log((7,NULL,"server shutdown initiated."));

    if (users!=NULL) cleanup_users(&users);
    if (roles!=NULL) cleanup_roles(&roles);

    shutdown_sources();
    shutdown_socket(input_socket,"from",offer_if);

    if (fetch_answer_s!=NULL) { sfree(fetch_answer_s); fetch_answer_s=NULL; };
    if (offer_if!=NULL) free_interface(offer_if);
    if (sb!=NULL) sb=free_socket_buf(sb);
    if (refinery_conn!=NULL) { free_interface(refinery_conn); refinery_conn=NULL; };
    while (my_outputs!=NULL) my_outputs=free_output(my_outputs);

    log((2,NULL,"*** server shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
