/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <asm/errno.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str/destination, provides interface */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */

/* ---------------------------------------------------------------------- */

socket_buf *new_socket_buf(int *socket)
{
    socket_buf *res=(socket_buf *)smalloc(sizeof(socket_buf));
    res->cur=0;
    res->cnt=0;
    res->avl=128;
    res->socket=socket;
    res->buf=(char*)smalloc(res->avl);
    return res;
}

socket_buf *free_socket_buf(socket_buf *sb)
{
    if (sb!=NULL)
    {
	if (sb->buf!=NULL) sfree(sb->buf); sb->buf=NULL;
	sfree(sb);
    };
    return NULL;
}

int socket_read_block(socket_buf *sb)
{
    char tmp[4096];
    int rc, res=true, sock=*(sb->socket);

    rc=read(sock,tmp,4096); if (rc==0) sleep(1);
    if (rc<0)
    {
	perror("read from main_socket");
	end=true;
	*(sb->socket)=-1;
	res=false;
    }
    else
    {
	int rest=(sb->avl)-(sb->cnt);
	while (rc>=rest)
	{
	    sb->avl=2*(sb->avl);
	    sb->buf=(char*)srealloc(sb->buf,sb->avl);
	    rest=(sb->avl)-(sb->cnt);
	};
	memcpy((sb->buf)+(sb->cnt),tmp,rc);
	sb->cnt = (sb->cnt)+rc;
    };
    if (2*(sb->cur) > (sb->avl))
    {
	memcpy((sb->buf),(sb->buf)+(sb->cur),(sb->cnt)-(sb->cur));
	sb->cnt=(sb->cnt)-(sb->cur);
	sb->cur=0;
    };
    return res;
}

int read_char(socket_buf *sb)
{
    int res=-1, ok=true;

    while ( (sb->cur>=sb->cnt) && (ok==true) ) ok=socket_read_block(sb);

    if (ok==true) { res=*((sb->buf)+(sb->cur)); (sb->cur)++; };

    return res;
}

int read_line(socket_buf *sb, str *s)
{
    int ch;
    char buf[2];

    buf[1]=0;
    reset_str(s);

    do
    {
	ch=read_char(sb);
	if (ch!=-1) { buf[0]=ch; add_str(s,buf); };
    }
    while ( (ch!=-1) && (ch!='\n') );

    return (ch==-1)?(false):(true);
}

void flush_buf(socket_buf *sb)
{
    sb->cur=0;
    sb->cnt=0;
}

/* ---------------------------------------------------------------------- */

int shutdown_socket(int sock, char *fromto, interface *i)
{
    if (sock!=-1)
    {
	log((2,NULL,"closing down stream %s '%s:%d'.",fromto,i->ip_address,i->port));
	close(sock);
    };
    return -1;
}

int re_initialise_input(interface *src)
{
    int res=-1;
    struct hostent *hp=gethostbyname(src->ip_address);
    if (hp==NULL)
    {
	static unsigned long gethostbyname_time=0;
	log((1,&gethostbyname_time,"error locating '%s' through gethostbyname().",src->ip_address));
	res=-1;
    }
    else
    {
	int sock=socket(AF_INET,SOCK_STREAM,0);
	if (sock<0)
	{
	    static unsigned long sock_time=0;
	    log((1,&sock_time,"error during socket() for %s:%d.",src->ip_address,src->port));
	    res=-1;
	}
	else
	{
	    struct sockaddr_in server;
	    server.sin_family=AF_INET;
	    memcpy(&(server.sin_addr),(hp->h_addr),hp->h_length);
	    server.sin_port=htons(src->port);
	    if (bind(sock,(struct sockaddr *)&server,sizeof(server))<0)
	    {
		static unsigned long bind_time=0;
		log((1,&bind_time,"error during bind(%s:%d).",src->ip_address,src->port));
		close(sock);
		res=-1;
	    }
	    else
	    {
		int l=listen(sock,5);
		if (l!=0)
		{
		    static unsigned long listen_time=0;
		    log((1,&listen_time,"error during listen() on %s:%d.",src->ip_address,src->port));
		    close(sock);
		    res=-1;
		}
		else
		{
		    /* the listener will cause a READABLE event
		       for SELECT.  It is possible that an ASYNC
		       NETWORK ERROR (see "man 2 select") occurs
		       before ACCEPT can be called.  To ensure
		       that ACCEPT never blocks, set the listener
		       to O_NONBLOCK */
		    if (fcntl(sock,F_SETFL,O_NONBLOCK)==-1)
		    {
			static unsigned long fcntl_time=0;
			log((1,&fcntl_time,"fcntl(%d,F_SETFL,O_NONBLOCK) for input socket failed.",sock));
			close(sock);
			res=-1;
		    }
		    else
		    {
			res=sock;
		    };
		};
	    };
	};
    };

    return res;
}

int re_initialise_output(interface *dst)
{
    int res=-1;
    struct hostent *hp=gethostbyname(dst->ip_address);
    if (hp==NULL)
    {
	static unsigned long gethostbyname_time=0;
	log((1,&gethostbyname_time,"error locating '%s' through gethostbyname().",dst->ip_address));
    }
    else
    {
	int sock=socket(AF_INET,SOCK_STREAM,0);
	if (sock<0)
	{
	    static unsigned long socket_time=0;
	    log((1,&socket_time,"error during socket() for %s:%d.",dst->ip_address,dst->port));
	}
	else
	{
	    struct sockaddr_in server;
	    server.sin_family=AF_INET;
	    memcpy(&(server.sin_addr),(hp->h_addr),hp->h_length);
	    server.sin_port=htons(dst->port);
	    if (connect(sock,(const struct sockaddr *)&server,sizeof(server))<0)
	    {
		static unsigned long connect_time=0;
		log((1,&connect_time,"error during connect(%s:%d).",dst->ip_address,dst->port));
		close(sock);
	    }
	    else
	    {
		log((2,NULL,"connected to %s:%d.",dst->ip_address,dst->port));
		res=sock;
	    };
	};
    };

    return res;
}

char *sockaddr_to_ascii(struct sockaddr_in *c)
{
    static char buf[50];
    unsigned long ip=ntohl(c->sin_addr.s_addr);
    sprintf(buf,"%lu.%lu.%lu.%lu:%d",
	    (ip>>24)&0xff,(ip>>16)&0xff,
	    (ip>>8)&0xff,ip&0xff,ntohs(c->sin_port));
    return buf;
}

/* ---------------------------------------------------------------------- */

source *current_sources=NULL;

source *new_source(source_t t, char *description)
{
    source *res=(source *)smalloc(sizeof(source));

    log((2,NULL,"creating new source '%s'.",description));

    res->tpe=t;
    res->src=-1;
    res->next=NULL;
    res->buffer[0]=0;

    res->state=sstate_unauthenticated;

    snprintf(res->description,100,"%s",description);

    switch(res->tpe)
    {
    case st_unknown:
	log((0,NULL,"unknown type in new_source()!!!\n"));
	exit(1);
	break;
    case st_watcher:
	res->t.watcher.pid=-1;
	res->t.watcher.frequency=0;
	res->t.watcher.name=NULL;
	res->t.watcher.command=NULL;
	res->t.watcher.parameter=NULL;
	break;
    case st_collector:
	break;
    case st_sender:
	break;
    case st_server:
	break;
    case st_viewer_1:
	break;
    };

    return res;
}

source *free_source(source *x)
{
    source *res=NULL;

    if (x!=NULL)
    {
	res=x->next;

	switch(x->tpe)
	{
	case st_unknown:
	    log((0,NULL,"unknown type in free_source()!!!\n"));
	    exit(1);
	    break;
	case st_watcher:
	    x->t.watcher.frequency=0;
	    if (x->t.watcher.pid!=-1)
	    {
		log((2,NULL,"sending sensor '%s' a KILL signal.",x->t.watcher.name));
		kill(x->t.watcher.pid,SIGKILL);
		waitpid(x->t.watcher.pid,NULL,WNOHANG);
		x->t.watcher.pid=-1;
	    };
	    if (x->t.watcher.name!=NULL)
	    { sfree(x->t.watcher.name); x->t.watcher.name=NULL; };
	    if (x->t.watcher.command!=NULL)
	    { sfree(x->t.watcher.command); x->t.watcher.command=NULL; };
	    if (x->t.watcher.parameter!=NULL)
	    { sfree(x->t.watcher.parameter); x->t.watcher.parameter=NULL; };
	    break;
	case st_collector:
	    break;
	case st_sender:
	    break;
	case st_server:
	    if (x->t.server.user!=NULL)
	    {
		sfree(x->t.server.user);
		x->t.server.user=NULL;
	    };
	    break;
	case st_viewer_1:
	    break;
	};

	if (x->src!=-1)
	{
	    log((2,NULL,"closing down source %s.",x->description));
	    close(x->src);
	    x->src=-1;
	};

	sfree(x);
    };

    return res;
}

int process_sources(int input_timeout,
		    int *accept_fd, struct sockaddr_in *client)
{
    fd_set rset;
    struct timeval timeout;
    int sel, max=0, res=-1;
    source *tmp=current_sources;

    if (current_sources==NULL)
    {
	log((7,NULL,"current_sources=NULL, sleep(%d).",input_timeout));
	sleep(input_timeout);
    };

    /* --- set one bit for each source --- */
    FD_ZERO(&rset);
    while (tmp!=NULL)
    {
	if (tmp->src!=-1)
	{
	    FD_SET(tmp->src,&rset);
	    if (tmp->src>max) max=tmp->src;
	};
	tmp=tmp->next;
    };

    /* --- set bit to also watch accept_fd --- */
    if (*accept_fd!=-1) FD_SET((*accept_fd),&rset);
    if (*accept_fd>max) max=*accept_fd;

    /* --- wait for input from any sensor --- */
    timeout.tv_usec=0;
    timeout.tv_sec=input_timeout;
    sel=select(max+1,&rset,NULL,NULL,&timeout);

    /* --- check accept_fd --- */
    if ( ((*accept_fd)>-1) && (sel>-1) )
    {
	int size=sizeof(*client);
	res=accept((*accept_fd), client, &size);
	if ( (res==-1) && (errno!=EAGAIN) )
	{
	    close(*accept_fd);
	    *accept_fd=-1;
	};
    };

    /* --- collect input from busy sensors --- */
    tmp=(sel==-1)?(NULL):(current_sources);
    while (tmp!=NULL)
    {
	if (tmp->src>-1)
	{
	    if (FD_ISSET(tmp->src,&rset))
	    {
		int r,l=strlen(tmp->buffer);
		if (l<BUFSIZE-1)
		{
		    r=read(tmp->src,tmp->buffer+l,BUFSIZE-2-l);
		    if (r>0)
		    {
			tmp->buffer[l+r]=0;
		    }
		    else
		    {
			if (r==0)
			{
			    static unsigned long read_time=0;
			    log((1,&read_time,"read 0 bytes from %d.",tmp->src));
			    tmp->src=-1;
			}
			else
			{
			    extern int errno;
			    if (errno!=EAGAIN)
			    {
				close(tmp->src);
				tmp->src=-1;
			    };
			};
		    };
		};
	    };
	};
	tmp=tmp->next;
    };

    return res;
}

void shutdown_sources()
{
    /* --- close down active ones --- */
    while (current_sources!=NULL)
	current_sources=free_source(current_sources);
}

/* ---------------------------------------------------------------------- */

void cleanup_sources()
{
    source **ptr=&current_sources;
    while (*ptr!=NULL)
    {
	source *tmp=*ptr;
	if (tmp->src==-1)
	{
	    log((0,NULL,"source '%s' has closed unexpectedly.",tmp->description));
	    *ptr=free_source(tmp);
	}
	else
	{
	    ptr = &((*ptr)->next);
	};
    };
}

/* ---------------------------------------------------------------------- */

void send_challenge(source *s)
{
    int i;
    char ch[18];
    for (i=0; i<8; i++) sprintf(s->challenge+2*i,"%02x",random()%256);
    sprintf(ch,"%s\n",s->challenge);
    log((7,NULL,"sending challenge '%s'.",s->challenge));
    write(s->src,ch,sizeof(ch));
}

char *authenticate_partner(char *line, source *s, user *u)
{
    int res=1;
    char *unme="somebody";

    while (u!=NULL)
    {
	char mytext[18];
	unsigned long i, len=16+1+strlen(u->name)+1+strlen(u->secret)+1;
	char *mysum, *expected=smalloc(len);
	memset(expected,0,len);
	sprintf(expected,"%s:%s:%s",u->name,u->secret,s->challenge);
	log((7,NULL,"authentication name/secret/challenge='%s' len=%lu",expected,len));
	mysum=mac_encrypt(expected,len);
	for (i=0; i<8; i++)
	    sprintf(mytext+2*i,"%02x",
		    ((unsigned char *)mysum)[i]);
	sprintf(mytext+16,"\n");
	log((7,NULL,"authentication hash should='%s' is='%s'",mytext,line));
	if (line!=NULL) line=index(line,':');
	if (line!=NULL) line++;
	if ( (line!=NULL) && (strcmp(mytext,line)==0) )
	{
	    unme=u->name;
	    res=0;
	    u=NULL;
	}
	else
	{
	    u=u->next;
	};
	sfree(expected);
    };

#ifdef NO_AUTHENTICATION
    fprintf(stderr,"ignoring authentication (res=%d)\n",res); res=0;
    log((0,NULL,"ignoring authentication!"));
#endif

    if (res==0)
    {
	log((2,NULL,"client has been authenticated successfully."));
	s->state=sstate_authenticated;
    }
    else
    {
	static unsigned long reject_time=0;
	log((1,&reject_time,"client has been rejected, closing socket."));
	close(s->src);
	(s->buffer)[0]=0;
	s->src=-1;
    };

    return sstrdup(unme);
}

/* ---------------------------------------------------------------------- */

int identify_self(int main_fd, user *me)
{
    unsigned char ch;
    str *resp=new_str();
    unsigned long i, len, cnt=0, avl=128;
    char mytext[18];
    char *mysum, *response, *challenge=(char*)smalloc(128);

    read(main_fd,&ch,1);
    while (ch!='\n')
    {
	if (cnt+1>=avl)
	{
	    avl=2*avl;
	    challenge=(char*)srealloc(challenge,avl);
	};
	challenge[cnt++]=ch;
	read(main_fd,&ch,1);
    };
    challenge[cnt]=0;
    log((7,NULL,"challenge='%s' len=%lu",challenge,cnt));
    len=cnt+1+strlen(me->name)+1+strlen(me->secret)+1;
    response=(char*)smalloc(len);
    sprintf(response,"%s:%s:%s",me->name,me->secret,challenge);
    log((7,NULL,"message='%s' len=%lu",response,len));
    mysum=mac_encrypt(response,len);
    for (i=0; i<8; i++)
	sprintf(mytext+2*i,"%02x",
		((unsigned char *)mysum)[i]);
    sprintf(mytext+16,"\n");
    add_str(resp,me->name);
    add_str(resp,":");
    add_str(resp,mytext);
    log((7,NULL,"response='%s'",cstr(resp)));
    write_string(resp,main_fd);
    free_str(resp);
    sfree(response);
    sfree(challenge);

    return main_fd;
}

/* ---------------------------------------------------------------------- */










