/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <time.h>
#include <ctype.h>
#include <stdio.h>

#include "log.h"
#include "tools.h"
#include "validate.h"

/* ---------------------------------------------------------------------- */

#define NAMELEN (VALIDATE_LABELLEN+1)
int name_cnt=0;
int name_avl=0;
char *namelist=NULL;

void reset_namelist()
{
    name_cnt=0;
    log((7,NULL,"reset_namelist()"));
}

int is_duplicate(char *name)
{
    int i=0, res=false;
    while (i<name_cnt)
    {
	if (strcmp(name,namelist+NAMELEN*i)==0)
	{
	    i=name_cnt;
	    res=true;
	};
	i++;
    };
    if (res==false)
    {
	while (name_cnt>=name_avl)
	{
	    name_avl=(name_avl==0)?(16):(2*name_avl);
	    namelist=(char*)srealloc(namelist,NAMELEN*name_avl);
	};
	strcpy(namelist+NAMELEN*name_cnt,name);
	name_cnt++;
    };
    return res;
}

/* ---------------------------------------------------------------------- */

char *eat_tab(char *buf, int *res)
{
    if (*buf=='\t')
    {
	buf++;
    }
    else
    {
	static unsigned long eattab_time=0;
	log((1,&eattab_time,"expected <TAB> instead of '%s'.",buf));
	*res=false;
    };

    return buf;
}

char *check_type(char *buf, char *txt, v_type tpe, v_type *res)
{
    int l=strlen(txt);
    if (memcmp(buf,txt,l)==0)
    {
	buf+=l;
	*res=tpe;
    };
    return buf;
}

char *eat_type(char *buf, v_type *tpe, int *res)
{
    char *errtmp=buf;
    v_type fnd=vt_unknown;

    if (fnd==vt_unknown) buf=check_type(buf,"int",vt_int,&fnd);
    if (fnd==vt_unknown) buf=check_type(buf,"float",vt_float,&fnd);
    if (fnd==vt_unknown) buf=check_type(buf,"timestamp",vt_timestamp,&fnd);
    if (fnd==vt_unknown) buf=check_type(buf,"text",vt_text,&fnd);

    if (*tpe==vt_unknown) *tpe=fnd;

    if (*tpe!=fnd)
    {
	static unsigned long eattype_time=0;
	log((1,&eattype_time,"expected <type %d> instead of <type %d> in '%s'.",*tpe,fnd,errtmp));
	*res=false;
    };

    return buf;
}

char *eat_title(char *buf, int *res)
{
    int count=0;
    char title_buffer[VALIDATE_LABELLEN+1];

    log((9,NULL,"eat_title(%s)",buf));

    while ( (isalnum(buf[count])) && (count<VALIDATE_LABELLEN) )
    {
	title_buffer[count]=buf[count];
	count++;
    };
    title_buffer[count]=0;

    if (count<1)
    {
	static unsigned long titlefield_time=0;
	log((1,&titlefield_time,"expected 'title' field, but got '%s'.",buf));
	*res=false;
    };

    if (count>VALIDATE_LABELLEN)
    {
	static unsigned long titlelen_time=0;
	log((1,&titlelen_time,"the 'title' field in a record may not exceed %d characters: '%s'.",VALIDATE_LABELLEN,buf));
	*res=false;
    };

    if (is_duplicate(title_buffer)==true)
    {
	static unsigned long titledup_time=0;
	log((1,&titledup_time,"a title ('%s') may not be used twice.",title_buffer));
	*res=false;
    };

    return buf+count;
}

char *eat_units(char *buf, int *res)
{
    int count=0;

    while (isalpha(buf[count])) count++;

    if (count<1)
    {
	static unsigned long unitsfield_time=0;
	log((1,&unitsfield_time,"expected 'units' field, but got '%s'.",buf));
	*res=false;
    };

    if (count>VALIDATE_LABELLEN)
    {
	static unsigned long unitslength_time=0;
	log((1,&unitslength_time,"the 'units' field in a record may not exceed %d characters: '%s'.",VALIDATE_LABELLEN,buf));
	*res=false;
    };

    return buf+count;
}

char *eat_int(char *buf, str *value, int *res)
{
    int count=0;

    while (isdigit(buf[count])) count++;

    if (count<1)
    {
	static unsigned long intfield_time=0;
	log((1,&intfield_time,"expected 'int' value, but got '%s'.",buf));
	*res=false;
    };

    if (count>VALIDATE_LABELLEN)
    {
	static unsigned long intlength_time=0;
	log((1,&intlength_time,"an 'int' value in a record may not exceed %d characters: '%s'.",VALIDATE_LABELLEN,buf));
	*res=false;
    };

    if (*res==true) cpy_str(value,buf,count);

    return buf+count;
}

char *eat_float(char *buf, str *value, int *res)
{
    int tmp, count=0;

    while (isdigit(buf[count]))	count++;

    if (count<1)
    {
	static unsigned long floatfield_time=0;
	log((1,&floatfield_time,"expected 'float' value, but got '%s'.",buf));
	*res=false;
    };

    if (buf[count]=='.')
    {
	count++;
    }
    else
    {
	static unsigned long floattypea_time=0;
	log((1,&floattypea_time,"expected 'float' value, but got '%s'.",buf));
	*res=false;
    };
    tmp=count;

    while (isdigit(buf[count])) count++;

    if ((count-tmp)<1)
    {
	static unsigned long floattypeb_time=0;
	log((1,&floattypeb_time,"expected 'float' value, but got '%s'.",buf));
	*res=false;
    };

    if (count>VALIDATE_LABELLEN)
    {
	static unsigned long floatlength_time=0;
	log((1,&floatlength_time,"a 'float' value in a record may not exceed %d characters: '%s'.",VALIDATE_LABELLEN,buf));
    };

    if (*res==true) cpy_str(value,buf,count);

    return buf+count;
}

char *eat_timestamp(char *buf, str *value, int *res)
{
    int count=0;

    while ( (count<22) && (*res==true) )
    {
	switch(count)
	{
	case  0: /* 4-digit year */
	case  1:
	case  2:
	case  3:
	    if (!isdigit(buf[count])) *res=false; break;
	case  4:
	    if (buf[count]!='-') *res=false; break;
	case  5: /* 2-digit month */
	case  6:
	    if (!isdigit(buf[count])) *res=false; break;
	case  7:
	    if (buf[count]!='-') *res=false; break;
	case  8: /* 2-digit month */
	case  9:
	    if (!isdigit(buf[count])) *res=false; break;
	case 10:
	    if (buf[count]!=' ') *res=false; break;
	case 11: /* 2-digit hour */
	case 12:
	    if (!isdigit(buf[count])) *res=false; break;
	case 13:
	    if (buf[count]!=':') *res=false; break;
	case 14: /* 2-digit minute */
	case 15:
	    if (!isdigit(buf[count])) *res=false; break;
	case 16:
	    if (buf[count]!=':') *res=false; break;
	case 17: /* 2-digit second */
	case 18:
	    if (!isdigit(buf[count])) *res=false; break;
	case 19:
	    if ( (buf[count]!='+') && (buf[count]!='-') ) *res=false; break;
	case 20: /* 2-digit timezone */
	case 21:
	    if (!isdigit(buf[count])) *res=false; break;
	default:
	    log((0,NULL,"internal timezone error."); break);
	};
	count++;
    };

    if (*res==false)
    {
	static unsigned long timetype_time=0;
	log((1,&timetype_time,"expected 'timestamp' value, but got '%s'.",buf));
    };

    if (*res==true) cpy_str(value,buf,count);

    return buf+count;
}

char *eat_text(char *buf, str *value, int *res)
{
    int count=0;

    while ( (buf[count]!='\t') && (buf[count]!='\n') )
	count++;

    if (count<1)
    {
	static unsigned long texttype_time=0;
	log((1,&texttype_time,"expected 'text' value, but got '%s'.",buf));
	*res=false;
    };

    if (count>VALIDATE_LABELLEN)
    {
	static unsigned long textlength_time=0;
	log((1,&textlength_time,"a 'text' value in a record may not exceed %d characters: '%s'.",VALIDATE_LABELLEN,buf));
    };

    if (*res==true) cpy_str(value,buf,count);

    return buf+count;
}

char *eat_value(char *buf, v_type tpe, str *value, int *res)
{
    static unsigned long datatype_time=0;

    reset_str(value);

    switch(tpe)
    {
    case vt_unknown:
	log((1,&datatype_time,"could not determine record's data type."));
	*res=false;
	break;
    case vt_int:
	buf=eat_int(buf,value,res);
	break;
    case vt_float:
	buf=eat_float(buf,value,res);
	break;
    case vt_timestamp:
	buf=eat_timestamp(buf,value,res);
	break;
    case vt_text:
	buf=eat_text(buf,value,res);
	break;
    };

    return buf;
}

/* ---------------------------------------------------------------------- */

char *eat_record(char *buf, v_type tpe, str *value, int *res)
{
    if (*res==true) buf=eat_tab(buf,res);
    if (*res==true) buf=eat_type(buf,&tpe,res);

    if (*res==true) buf=eat_tab(buf,res);
    if (*res==true) buf=eat_title(buf,res);

    if (*res==true) buf=eat_tab(buf,res);
    if (*res==true) buf=eat_units(buf,res);

    if (*res==true) buf=eat_tab(buf,res);
    if (*res==true) buf=eat_value(buf,tpe,value,res);

    return buf;
}

/* ---------------------------------------------------------------------- */

str *validate_line_value=NULL;
#define value validate_line_value
int validate_line(char *buf, str *site, str *machine, str *sensor, str *ts)
{
    int res=true;
    if (value==NULL) { value=new_str(); } else { reset_str(value); };

    reset_namelist();

    if (res==true) buf=eat_record(buf,vt_text,site,&res);
    if (res==true) buf=eat_record(buf,vt_text,machine,&res);
    if (res==true) buf=eat_record(buf,vt_text,sensor,&res);
    if (res==true) buf=eat_record(buf,vt_timestamp,ts,&res);

    while ( (res==true) && (*buf!='\n') )
	buf=eat_record(buf,vt_unknown,value,&res);

    return res;
}
#undef value

/* ---------------------------------------------------------------------- */

void validate_cleanup()
{
    if (validate_line_value!=NULL)
    { free_str(validate_line_value); validate_line_value=NULL; };
    if (namelist!=NULL)
    { sfree(namelist); namelist=NULL; name_cnt=0; name_avl=0; };
}

/* ---------------------------------------------------------------------- */

