/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str/md_lst() */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str, provides interface/out_struct */
#include "db.h" /* provides tables */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "output.h" /* needs str/tables/out_struct/socket_buf */
#include "cleanup.h"

/* ---------------------------------------------------------------------- */

int my_outputs_cnt=0;
out_struct *my_outputs=NULL;
out_struct *timed_outputs(socket_buf *sb, int *ocnt, unsigned long freq)
{
    out_struct *res;
    static time_t last=0;
    time_t now=time(NULL);

    if ( (my_outputs==NULL) || (now-last<freq) )
    {
	last=now;
	while (my_outputs!=NULL) my_outputs=free_output(my_outputs);
	my_outputs=get_outputs(sb, &my_outputs_cnt);
    };

    res=my_outputs;
    *ocnt=my_outputs_cnt;

    return res;
}

/* ---------------------------------------------------------------------- */

out_struct *find_output(out_struct *outputs, char *request)
{
    out_struct *res=NULL;
    while (outputs!=NULL) if (strcmp(outputs->name,request)==0)
    {
	res=outputs;
	outputs=NULL;
    }
    else
    {
	outputs=outputs->next;
    };

    return res;
}

/* ---------------------------------------------------------------------- */

int fetch_output_timed(socket_buf *sb, out_struct *asked, str *s)
{
    int res=false;
    unsigned long ts=time(NULL);

    log((7,NULL,"fetch_output_timed(%s)",asked->name));

    if ( (ts-asked->data_timestamp>asked->delay) || (asked->data==NULL) )
	fetch_output(sb,asked);

    reset_str(s);
    if (asked->data!=NULL)
    {
	res=true;
	add_str(s,asked->data);
    };

    return res;
}

/* ---------------------------------------------------------------------- */

static str *fetch_answer_s=NULL;
#define s fetch_answer_s
/* this must check the syntax of the data block! */
char *fetch_answer(char *request, socket_buf *sb, unsigned long data_freq)
{
    int ocnt=0;
    unsigned long dir_freq=data_freq;
    out_struct *outputs=timed_outputs(sb,&ocnt,dir_freq);

    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    if (outputs==NULL)
    {
	add_str(s,"\tdirectory\t0\n");
    }
    else
    {
	int do_dir=false;
	out_struct *asked=find_output(outputs,request);
	if (asked==NULL)
	{
	    do_dir=true;
	}
	else
	{
	    if (fetch_output_timed(sb,asked,s)==false) do_dir=true;
	};
    };

    return cstr(s);
}
#undef s

/* ---------------------------------------------------------------------- */

int process_command(char *cmd, char *user, int socket, socket_buf *sb, unsigned long freq)
{
    int res=socket;
    str *s=new_str();
    char *request=NULL, *answer=NULL;
    unsigned long count=0;

    if (cmd!=NULL) if (cmd[0]=='\t') request=cmd+1;
    if (request!=NULL)
    {
	cmd=index(request,'\t');
	if (cmd!=NULL) *cmd++=0;
	if (cmd!=NULL) sscanf(cmd,"%lu",&count);
    };

    answer=
	(request==NULL)
	?fetch_answer("xxx",sb,freq)
	:fetch_answer(request,sb,freq);

    write_string(s,socket);
    free_str(s);

    return res;
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    char rstate[128];
    user *identity=NULL;
    char *log_file="viewer-1.log";
    interface *server_conn=NULL;
    char *config_file="viewer-1.conf";
    int input_timeout=CPU_SAVER;
    unsigned long data_freq=30;
    viewer_input *viewer_inputs=NULL;
    socket_buf *sb=new_socket_buf(&main_fd);

    initstate(time(NULL),rstate,128);

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("viewer-1");

    log((2,NULL,"*** viewer-1 startup ***"));

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_viewer_1);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    cleanup_users(&identity); identity=c->identity; c->identity=NULL;
	    cleanup_viewer_inputs(&viewer_inputs); viewer_inputs=c->viewer_inputs; c->viewer_inputs=NULL;

	    if ( (server_conn==NULL) ||
		 ( (server_conn->port!=c->src->port) ||
		   (strcmp(server_conn->ip_address,c->src->ip_address)!=0) ) )
	    {
		if (server_conn!=NULL) free_interface(server_conn);
		server_conn=new_interface(c->src->ip_address,c->src->port);
		if (main_fd!=-1)
		    main_fd=shutdown_socket(main_fd,"to",server_conn);
	    };

	    data_freq=c->data_freq;
	    input_timeout=c->input_timeout;

	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	log((7,NULL,"main_fd=%d.",main_fd));
	if (server_conn!=NULL)
	    log((7,NULL,"server_conn=%s:%d.",server_conn->ip_address,server_conn->port));
	if (identity!=NULL)
	    log((7,NULL,"identity=%s/%s.",identity->name,identity->secret));
	if ( (main_fd==-1) && (server_conn!=NULL) && (identity!=NULL) )
	{
	    log((7,NULL,"attempting to open connection to %s:%d.",server_conn->ip_address,server_conn->port));

	    main_fd=re_initialise_output(server_conn);
	    if (main_fd!=-1)
	    {
		log((7,NULL,"success at opening connection to %s:%d.",server_conn->ip_address,server_conn->port));
		main_fd=identify_self(main_fd,identity);
	    };
	};

	if (main_fd!=-1)
	{
	    viewer_input *v=viewer_inputs;
	    while (v!=NULL)
	    {
		char *data_name=v->name;
		char *data_file=v->file;
		unsigned long rows=v->rows;
		char *answer=fetch_answer(data_name,sb,data_freq);

		if (strncmp("\tdirectory\t",answer,11)==0)
		{
		    /* do nothing */
		}
		else
		{
		    FILE *d;
		    char buf[1000];

		    sprintf(buf,"%s.tmp",data_file);
		    d=fopen(buf,"w");
		    if (d!=NULL)
		    {
			int len=strlen(answer);
			char *line=extract_first_line(answer,&len);
			char *num=index(index(line,'\t')+1,'\t')+1;
			unsigned long i, count=atoi(num);
			count=(count>rows)?(rows):(count);

			for (i=0; i<count; i++)
			{
			    line=extract_first_line(answer,&len);
			    if ( (line!=NULL) && (*line=='\t') ) line++;
			    if (line!=NULL) fprintf(d,"%s",line);
			};
			fclose(d);
			rename(buf,data_file);
		    };
		};
		v=v->next;
	    };
	};

	sleep(data_freq);
    };

    log((7,NULL,"viewer_1 shutdown initiated."));

    shutdown_sources();

    cleanup_viewer_inputs(&viewer_inputs);

    if (fetch_answer_s!=NULL) { sfree(fetch_answer_s); fetch_answer_s=NULL; };
    if (sb!=NULL) sb=free_socket_buf(sb);
    if (server_conn!=NULL) { free_interface(server_conn); server_conn=NULL; };

    while (my_outputs!=NULL) my_outputs=free_output(my_outputs);

    log((2,NULL,"*** viewer_1 shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
