/* -*- Mode: C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <asm/errno.h>
#include <netinet/in.h>

#include "tools.h" /* provides str/md_lst() */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str, provides interface/out_struct */
#include "db.h" /* provides tables */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "output.h" /* needs str/tables/out_struct/socket_buf */
#include "cleanup.h"

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    char rstate[128];
    user *identity=NULL;
    char *log_file="viewer-2.log";
    interface *server_conn=NULL;
    char *config_file="viewer-2.conf";
    int input_timeout=CPU_SAVER;
    viewer_input *viewer_inputs=NULL;
    socket_buf *sb=new_socket_buf(&main_fd);

    initstate(time(NULL),rstate,128);

    parse_commands(argc, argv, &config_file, &log_file);

    log_setfile(log_file);
    log_setapp("viewer-2");

    log((2,NULL,"*** viewer-2 startup ***"));

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_viewer_2);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    cleanup_users(&identity); identity=c->identity; c->identity=NULL;

	    if ( (server_conn==NULL) ||
		 ( (server_conn->port!=c->src->port) ||
		   (strcmp(server_conn->ip_address,c->src->ip_address)!=0) ) )
	    {
		if (server_conn!=NULL) free_interface(server_conn);
		server_conn=new_interface(c->src->ip_address,c->src->port);
		if (main_fd!=-1)
		    main_fd=shutdown_socket(main_fd,"to",server_conn);
	    };

	    input_timeout=c->input_timeout;

	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	log((7,NULL,"main_fd=%d.",main_fd));
	if (server_conn!=NULL)
	    log((7,NULL,"server_conn=%s:%d.",server_conn->ip_address,server_conn->port));
	if (identity!=NULL)
	    log((7,NULL,"identity=%s/%s.",identity->name,identity->secret));
	if ( (main_fd==-1) && (server_conn!=NULL) && (identity!=NULL) )
	{
	    log((7,NULL,"attempting to open connection to %s:%d.",server_conn->ip_address,server_conn->port));

	    main_fd=re_initialise_output(server_conn);
	    if (main_fd!=-1)
	    {
		log((7,NULL,"success at opening connection to %s:%d.",server_conn->ip_address,server_conn->port));
		main_fd=identify_self(main_fd,identity);
	    };
	};

	if (main_fd!=-1)
	{
	    int cnt=0;
	    str *s=new_str();
	    out_struct *outputs=get_outputs(sb, &cnt);
	    produce_directory(outputs,cnt,s,NULL,NULL,false);
	    printf("%s",cstr(s));
	    while (outputs!=NULL) outputs=free_output(outputs);
	    free_str(s);
	    end=1;
	};
    };

    log((7,NULL,"viewer-2 shutdown initiated."));

    shutdown_sources();

    cleanup_viewer_inputs(&viewer_inputs);

    if (sb!=NULL) sb=free_socket_buf(sb);
    if (server_conn!=NULL) { free_interface(server_conn); server_conn=NULL; };

    log((2,NULL,"*** viewer-2 shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
