/* -*- Mode:C++ -*- */

/* ---------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <asm/errno.h>

#include "tools.h" /* provides str */
#include "destination.h" /* provides destination */
#include "conf-tools.h" /* needs str/destination, provides interface */

#include "log.h"
#include "cache.h"
#include "source.h" /* needs interface */
#include "cleanup.h"

/* ---------------------------------------------------------------------- */

/* find sensor l in list old, extract if found */
source *extract_source(sensor *l, source **list)
{
    source *res=NULL;
    source **entry=list;

    while (*entry!=NULL)
    {
	source *old=*entry;
	if ( (strcmp(l->name,old->t.watcher.name)==0) &&
	     (strcmp(l->command,old->t.watcher.command)==0) &&
	     (strcmp(l->parameter,old->t.watcher.parameter)==0) &&
	     (l->frequency==old->t.watcher.frequency) )
	{
	    res=old;
	    *entry=old->next;
	    while (*entry!=NULL) entry=&((*entry)->next);
	}
	else
	{
	    entry=&((*entry)->next);
	};
    };

    return res;
}

/* ---------------------------------------------------------------------- */

void prepare_sources(config *c)
{
    int count=1;
    source *old_sources=current_sources;
    current_sources=NULL;

    while (c->sensor_list!=NULL)
    {
	source *x=extract_source(c->sensor_list,&old_sources);
	if (x==NULL)
	{
	    x=new_source(st_watcher,c->sensor_list->name);

#define SETS(X) x->t.watcher.X=sstrdup(c->sensor_list->X);
	    SETS(name);
	    SETS(command);
	    SETS(parameter);

#define SETI(X) x->t.watcher.X=c->sensor_list->X;
	    SETI(frequency);
	}
	else
	{
	    log((7,NULL,"re-using source '%s'.",x->t.watcher.name));
	};
	if (count>SOURCES)
	{
	    static unsigned long sourcecnt_time = 0;
	    log((1,&sourcecnt_time,"more than %d sources, ignoring '%s'.",SOURCES,x->t.watcher.name));
	    free_source(x);
	}
	else
	{
	    x->next=current_sources;
	    current_sources=x;
	};
	c->sensor_list=free_sensor(c->sensor_list);
	count++;
    };

    while (old_sources!=NULL)
	old_sources=free_source(old_sources);
}

int flush_output(interface *destination, int sock)
{
    char *buf=cache_extract();

    while ( (sock>-1) && (buf!=NULL) )
    {
	int length=strlen(buf), written=0;
	while ( (sock>-1) && (written!=length) )
	{
	    int rc=write(sock,buf+written,length-written);
	    if (rc>-1)
	    {
		written+=rc;
	    }
	    else
	    {
		if (end==false)
		{
		    static unsigned long outwrite_time = 0;
		    log((1,&outwrite_time,"error during write(%d) to output stream.",sock));
		};
		cache_insert(buf);
		if (rc!=EAGAIN)
		{
		    log((2,NULL,"shutting down output stream to '%s:%d'.",destination->ip_address,destination->port));
		    close(sock);
		    sock=-1;
		};
	    };
	};

	if (written==length)
	    buf=cache_extract();
    };

    return sock;
}

int child_pipe(char *cmd, char *freq, char *param, pid_t *pid)
{
    int out[2], res=-1;

    if (pipe(out)!=0)
    {
	log((0,NULL,"error creating pipe."));
    }
    else
    {
	pid_t p=fork();
	if (p==(pid_t)-1)
	{
	    log((0,NULL,"error during fork()."));
	    close(out[0]);
	    close(out[1]);
	    *pid=-1;
	}
	else
	{
	    if (p!=0)
	    {
		/* reader/parent process */
		/* close unused file descriptor */
		close(out[1]);

		res=out[0];
		if (res==-1)
		{
		    int status=0;
		    kill(p,SIGKILL);
		    waitpid(p,&status,0);
		    log((0,NULL,"error opening out[0] for reading."));
		    *pid=-1;
		}
		else
		{
		    *pid=p;
		};
	    }
	    else
	    {
		/* child process */
		/* close unused file descriptors */
		source *tmp=current_sources;
		while (tmp!=NULL)
		{
		    close(tmp->src);
		    tmp=tmp->next;
		};
		close(0); close(1); close(2); close(out[0]);

		/* use parent descriptor */
		if (dup2(out[1],1) == -1)
		{
		    log((0,NULL,"error during dup2(x,1)."));
		    exit(1);
		}
		else
		{
		    extern char **environ;
		    char *argv[] = { cmd, freq, param };
		    if (setsid()==-1)
			log((0,NULL,"setsid failed for sensor '%s'.",cmd));
		    execve(cmd,argv,environ);
		    log((0,NULL,"execve of '%s' failed.",cmd));
		    exit(1);
		};
	    };
	};
    };

    return res;
}

static str *produce_line_s=NULL;
#define s produce_line_s
str *produce_line(char *machine, char *sensor, char *now, char *buffer)
{
    str *res=NULL;
    int l=strlen(buffer);
    char *line=extract_last_line(buffer,&l);
    if (s==NULL) { s=new_str(); } else { reset_str(s); };

    if (line!=NULL)
    {
	add_str(s,"\ttext\tmachine\tname\t"); add_str(s,machine);
	add_str(s,"\ttext\tsensor\tname\t"); add_str(s,sensor);
	add_str(s,"\ttimestamp\ttimestamp\tepoch\t"); add_str(s,now);
	add_str(s,line);
	res=s;
    };

    return res;
}
#undef s

/* ---------------------------------------------------------------------- */

void re_initialise_pipes()
{
    source *tmp=current_sources;

    /* --- open one pipe for each sensor --- */
    while (tmp!=NULL)
    {
	if (tmp->src==-1)
	{
	    char freq[10];
	    sprintf(freq,"%d",tmp->t.watcher.frequency);

	    log((7,NULL,"opening sensor '%s' with command '%s %d%s%s'.",tmp->t.watcher.name,tmp->t.watcher.command,tmp->t.watcher.frequency,(strlen(tmp->t.watcher.parameter)==0)?(""):(" "),tmp->t.watcher.parameter));
	    tmp->src=child_pipe(tmp->t.watcher.command,freq,tmp->t.watcher.parameter,&(tmp->t.watcher.pid));
	    if (tmp->src>-1)
	    {
		log((2,NULL,"opening sensor '%s' with command '%s' succeeded.",tmp->t.watcher.name,tmp->t.watcher.command));
	    }
	    else
	    {
		static unsigned long sensoropen_time = 0;
		log((1,&sensoropen_time,"opening sensor '%s' with command '%s' failed.",tmp->t.watcher.name,tmp->t.watcher.command));
	    };
	};
	tmp=tmp->next;
    };
}

void process_watcher_sources(char *machine_name)
{
    /* --- process any complete input line --- */
    if (machine_name!=NULL)
    {
	char *now=iso_timestamp();
	source *tmp=current_sources;
	while (tmp!=NULL)
	{
	    if ((tmp->buffer)[0]!=0)
	    {
		str *res;
		do
		{
		    res=produce_line(machine_name,
				     tmp->t.watcher.name,
				     now,
				     tmp->buffer);
		    if (res!=NULL)
		    {
			cache_insert(cstr(res));
			/* produce line has a local static string */
		    };
		}
		while (res!=NULL);
	    };
	    tmp=tmp->next;
	};
    };
}

/* ---------------------------------------------------------------------- */

int main(int argc, char **argv)
{
    user *identity=NULL;
    char *log_file="watcher.log";
    char *machine_name=NULL;
    char *config_file="watcher.conf";
    interface *destination=NULL;
    int input_timeout=CPU_SAVER, input_socket=-1;

    parse_commands(argc, argv, &config_file, &log_file);

    log_setapp("watcher");
    log_setfile(log_file);

    log((2,NULL,"*** watcher startup ***"));

    install_handler();

    while (end==false)
    {
	config *c=fetch_config(config_file,ct_watcher);
	if (c!=NULL)
	{
	    log_setfile(c->log_file);
	    log_setlevel(c->log_level);

	    input_timeout=c->input_timeout;
	    cleanup_users(&identity); identity=c->identity; c->identity=NULL;
	    if (machine_name!=NULL) sfree(machine_name);
	    machine_name=sstrdup(c->machine_name);
	    if ( (destination==NULL) ||
		 ( (destination->port!=c->dst->port) ||
		   (strcmp(destination->ip_address,c->dst->ip_address)!=0) ) )
	    {
		if (destination!=NULL) free_interface(destination);
		destination=new_interface(c->dst->ip_address,c->dst->port);
		main_fd=shutdown_socket(main_fd,"to",destination);
	    };
	    prepare_sources(c);
	    free_config(c);
	    log((2,NULL,"read new config file."));
	};

	log((7,NULL,"main_fd=%d.",main_fd));

	if (main_fd==-1)
	{
	    if (destination!=NULL)
	    {
		log((7,NULL,"attempting to open connection to %s:%d.",destination->ip_address,destination->port));

		main_fd=re_initialise_output(destination);
		if (main_fd!=-1)
		{
		    log((7,NULL,"success at opening connection to %s:%d.",destination->ip_address,destination->port));
		    main_fd=identify_self(main_fd,identity);
		};
	    };
	}
	else
	{
	    main_fd=flush_output(destination,main_fd);
	};

	if (machine_name!=NULL)
	{
	    re_initialise_pipes();
	    (void)process_sources(input_timeout,&input_socket,NULL);
	    process_watcher_sources(machine_name);
	};

	if ( (main_fd==-1) && (current_sources==NULL) ) sleep(CPU_SAVER);
    };

    log((7,NULL,"watcher shutdown initiated."));

    shutdown_sources();
    shutdown_socket(main_fd,"to",destination);

    if (identity!=NULL) cleanup_users(&identity);
    if (machine_name!=NULL) sfree(machine_name); machine_name=NULL;
    if (destination!=NULL) free_interface(destination); destination=NULL;
    if (produce_line_s!=NULL) free_str(produce_line_s); produce_line_s=NULL;

    log((2,NULL,"*** watcher shutdown ***"));

    /* the logging module is cleaning up its data here ... */
    do_cleanup();

    md_lst(__FILE__,__LINE__);

    return 0;
}

/* ---------------------------------------------------------------------- */
